/// t2c_req.h
/// REQ-related facilities - declarations. 
/// Note that REQ-macros themselves are not defined in this file - only  
/// the stuff that does not depend on the test execution framework (TET, 
/// xUnit, ...) is here.

/////////////////////////////////////////////////////////////////////////////
// Copyright 2009-2010 
// Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//    http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License. 
/////////////////////////////////////////////////////////////////////////////

#ifndef T2C_REQ_H_1102_INCLUDED
#define T2C_REQ_H_1102_INCLUDED

#include <t2c_rcat.h>
#include <t2c_util.h>

#ifdef __cplusplus
extern "C" {
#endif

/////////////////////////////////////////////////////////////////////////////
// Defines
/////////////////////////////////////////////////////////////////////////////

/// Maximum length of a prefix (in characters) in a requirement ID (without dots).
#define T2C_MAX_PREFIX_LEN  10

/////////////////////////////////////////////////////////////////////////////
// Functions
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// [NB] Individual IDs in 'id_str' arguments can be separated by any number of
// spaces, tabs, semicolons and commas.

/// Return 0 if the requirements specified in 'id_str' should not be checked, 
/// nonzero otherwise.
/// If 'check_ext' is nonzero, "ext"-requirements need to be checked, and vice
/// versa.
/// If 'check_ext' is nonzero, the function returns nonzero (i.e. "need to check").
/// If 'check_ext' is 0 but there is at least one requirement ID in 'id_str' 
/// with "ext" prefix, the function returns zero ("should not check").
/// Otherwise the function returns nonzero.
int
t2c_req_need_check(const char* id_str, int check_ext);

/// Handle failed requirements with IDs specified in 'id_str' and return final 
/// verdict: T2C_TR_UNRESOLVED for violated "app"-requirements, T2C_TR_FAILED
/// for violated requirements for the system under test.
/// The caller must ensure that "app" and non-"app" requirements are not mixed
/// in 'id_str' otherwise the behaviour of the function is undefined. At least,
/// the function will not crash.
/// Mixing requirements of these kinds makes no sense anyway.
/// 
/// If 'id_str' is NULL, empty string or if it consists entirely of separator 
/// characters, the behaviour of the function is undefined. At least, it will not
/// crash.
///
/// If 'msg' is not NULL or an empty string it is output first (via 'print_func')
/// before any other output from this function.
/// 
/// Requirement catalogue 'rc' is used to lookup the text of failed requirements.
/// If the requirement is not found in the catalogue, "<<< REQ NOT FOUND >>>"
/// is used.
///
/// The function uses 'print_func' to output messages (IDs, requirement text, etc.)
/// to the logging facility supported by 'print_func'
/// 
/// In case of an internal error (out of memory or failure reported by 'print_func')
/// the function may return T2C_TR_ERROR.
ET2CTestResult
t2c_req_handle_failure(const char* id_str, const char* msg, CReqCat* rc,
    FT2CPrintFunc print_func);

/// Output the IDs of the checked requirements (all IDs from 'id_str') in 
/// an implementation-defined format.
/// 'print_func' is used to do the output.
/// The function returns nonzero if successful, 0 otherwise.
int
t2c_req_print_checked(const char* id_str, FT2CPrintFunc print_func);

/////////////////////////////////////////////////////////////////////////////

#ifdef __cplusplus
}
#endif

#endif // T2C_REQ_H_1102_INCLUDED
