/// t2c_util.h
/// Utility stuff provided by T2C API - declarations.

/////////////////////////////////////////////////////////////////////////////
// Copyright 2009-2010 
// Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//    http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License. 
/////////////////////////////////////////////////////////////////////////////

#ifndef T2C_UTIL_H_1100_INCLUDED
#define T2C_UTIL_H_1100_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

/////////////////////////////////////////////////////////////////////////////
/// ETestResult
/// The result codes that test functions, t2c_spawn() and other functions can
/// return. Note that the test functions can only return codes from 
/// T2C_TR_PASS to T2C_TR_FIP inclusive.
/// T2C_TR_UNINITIATED and T2C_TR_TIME_EXPIRED can be returned by t2c_spawn()
/// but not by the test functions. 
/// T2C_TR_ERROR can be returned by t2c_req_handle_failure(), for example.
///
/// [NB] In some targets, each test can be executed in a separate process
/// and the result of the test is determined by the exit code of that process
/// (if it finishes normally) or from the process status information (in case
/// of abnormal termination).
/// That is, if exit(<code>) is called in the test, <code> will be considered
/// the result code of the test. For example, some functions in GTK and related
/// libraries may call exit(1) if the operation they perform fails (gtk_init, 
/// etc.). 1 (EXIT_FAILURE - see stdlib.h) is mapped to T2C_TR_UNRESOLVED here
/// which seems reasonable.
///
/// Anyway, it is recommended not to use exit() in the tests explicitly.
/////////////////////////////////////////////////////////////////////////////
typedef enum ET2CTestResult_
{
    /// Some internal error occured, probably due to the lack of memory or
    /// failure to output message to the logging facility being used, etc.
    T2C_TR_ERROR        = -1,
    
    /// Test passed.
    T2C_TR_PASSED       = 0,
    
    /// Some error occured during test execution, may be not related to
    /// the functionality being tested. 
    /// Anyway, not sure whether the test would pass or fail if the error
    /// did not occur. 
    T2C_TR_UNRESOLVED   = 1, 
    
    /// Test failed: the system under test definitely did something wrong.
    T2C_TR_FAILED       = 2,    
        
    /// The functionality to be tested is optional and it is not supported
    /// by the particular system under test.
    T2C_TR_UNSUPPORTED  = 3,
    
    /// The mandatory functionality to be tested is not tested for some reason.
    T2C_TR_UNTESTED     = 4,
    
    /// The test has not been even initiated due to some failure in a test
    /// startup function (see FTestStartupFunc in t2c_spawn.h and do not 
    /// confuse it with the global startup function that is created from
    /// <STARTUP> section).
    T2C_TR_UNINITIATED  = 5,
    
    /// This result code is set by t2c_spawn() if a test function executed
    /// in a child process did not finish before the wait time ran out.
    /// See description of <$WAIT_TIME$> parameter for details.
    T2C_TR_TIME_EXPIRED = 65,
        
    /// "FIP" means "For Information Provider". 
    /// That means, the functionality has been tested but the results of 
    /// the test (e.g. saved execution traces, generated images, etc.) 
    /// need to be reviewed to decide whether the test passed or not.
    T2C_TR_FIP          = 102
} ET2CTestResult;

/////////////////////////////////////////////////////////////////////////////
// Other types
/////////////////////////////////////////////////////////////////////////////

/// Type of a function that outputs the message (its argument) somewhere.
/// Like printf, the function should return negative value on failure and
/// nonnegative value otherwise.
/// Unlike printf, it is not required that the function returns the number of 
/// characters transfered to destination (log file, stderr, ...) on success.
/// Unlike printf, this function is supposed to have only one argument rather
/// than a variable number of them.
typedef int (*FT2CPrintFunc)(const char*);

/////////////////////////////////////////////////////////////////////////////
// Functions
/////////////////////////////////////////////////////////////////////////////

/// Construct the path to the specified test data and returns the resulting
/// string. NULL is returned if there is not enough memory to perform the 
/// operation.
/// The test data directory is <suite_root_dir>/testdata/<test_name>.
/// 'rel_path' is a path to the data relative to this directory.
/// The function does not check if the resulting path or any of its parts exists.
///
/// The returned pointer should be freed by the caller when no longer needed.
char*
t2c_get_data_path(const char* suite_root_dir, const char* test_name, 
                  const char* rel_path);

/// Set new result code to 'rc' taking its previous value and precedence of 
/// the codes into account.
/// The precedence order (highest first) is as follows:
///   T2C_TR_FAILED
///   T2C_TR_UNRESOLVED, T2C_TR_UNINITIATED
///   T2C_TR_TIME_EXPIRED, T2C_TR_FIP
///   T2C_TR_UNSUPPORTED, T2C_TR_UNTESTED
///   T2C_TR_PASSED
/// If 'new_rc' has higher precedence than current value of '*rc', 'new_rc' 
/// code is set. Otherwise '*rc' remains unchanged.
void
t2c_set_result_code(ET2CTestResult* rc, ET2CTestResult new_rc);

/// Convert the specified string to a number and return the latter.
/// If 'str' is an empty string, the function returns 'prev'.
/// If an error occurs or if the function is able to convert only a part of 
/// the string, the function issues a warning using print_func() and 
/// returns 'prev'.
/// 'test_name' will be used in that warning output.
/// [NB] One common use case of the function is as follows (in a template):
///     some_param = t2c_get_number("<$SOME_PARAM$>", (long)some_param);
/// That is, if "SOME_PARAM" parameter is defined and is not an empty string,
/// the value of 'some_param' variable will be set from "SOME_PARAM".
/// Otherwise 'some_param' will remain unchanged.
long
t2c_get_number(const char* str, long prev, const char* test_name, 
    FT2CPrintFunc print_func);

/// Return an absolute path to the directory containing the file specified by
/// 'file_path' (the latter may be relative or absolute). 
/// The function does not check whether the path exists in the filesystem.
/// The function returns NULL in case of any error (out of memory, etc.).
/// The returned pointer must be freed when no longer needed.
char* 
t2c_get_dir(const char* file_path);

/// Create a new string containing the absolute path corresponding to the 
/// concatenation of 'base_path' and 'rel_path' (with '/' added in between 
/// if necessary). 
/// 'base_path' may be relative too. If so, it is considered relative to the 
/// current directory.
/// 
/// NULL is returned if there is not enough memory to perform the operation.
/// The returned pointer should be freed by the caller when no longer needed.
char*
t2c_path_sum(const char* base_path, const char* rel_path);

#ifdef __cplusplus
}
#endif

#endif // T2C_UTIL_H_1100_INCLUDED
