/// t2c_req.c
/// REQ-related facilities - implementation.

/////////////////////////////////////////////////////////////////////////////
// Copyright 2009-2010 
// Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//    http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License. 
/////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <smap.h>
#include <grar.h>
#include <mist_string_utils.h>
#include <t2c_req.h>

///////////////////////////////////////////////////////////////////////////
// Private variables
///////////////////////////////////////////////////////////////////////////

// Separator characters that can be used in the list of requirement IDs.
static const char seps[] = " \t;,";
static const size_t sz_seps = sizeof(seps) / sizeof(seps[0]) - 1;

///////////////////////////////////////////////////////////////////////////
// Declaration: private methods
///////////////////////////////////////////////////////////////////////////

/// Split the list of requirement IDs 'id_str' into separate requirements,
/// initialize 'reqs' array via grar_create and populate it with pointers to
/// the newly allocated copies of these IDs.
/// Use grar_destroy_with_elements() to free both the array and its contents.
/// The function returns nonzero if successful, 0 otherwise (if there is not
/// enough memory to complete the operation).
/// If the function fails, 'reqs' will remain uninitialized.
/// If 'id_str' empty or contains only separator characters, it is not considered
/// an error. The resulting array will contain one element in this case - 
/// a pointer to empty string.
int
t2c_req_split_list(const char* id_str, CGrowingArray* reqs);

/// Return nonzero if at least one of the IDs in 'id_str' has specified 
/// prefix, 0 otherwise.
/// Note that an ID can have various prefixes separated by single dots,
/// so the prefix to look for may be not the first one in a given ID.
/// 'prefix' must contain no dots and must not be empty. Its length must not
/// exceed T2C_MAX_PREFIX_LEN.
int
t2c_req_has_prefixed_req(const char* id_str, const char* prefix); 

///////////////////////////////////////////////////////////////////////////
// Implementation: public methods
///////////////////////////////////////////////////////////////////////////

int
t2c_req_need_check(const char* id_str, int check_ext)
{
    assert(id_str != NULL);
    
    if (check_ext)
    {
        return 1;
    }
    
    return !t2c_req_has_prefixed_req(id_str, "ext");
}

int
t2c_req_print_checked(const char* id_str, FT2CPrintFunc print_func)
{
    assert(id_str != NULL);
    assert(print_func != NULL);
    
    static const char* fmt = "Checked requirement: {%s}";
    char* buf = (char*)malloc(strlen(fmt) + strlen(id_str) + 1); // should be more than enough
    
    if (buf == NULL)
    {
        return 0;
    }
    
    CGrowingArray reqs;
    if (!t2c_req_split_list(id_str, &reqs))
    {
        free(buf);
        return 0;
    }
    
    char** ids = grar_get_c_array(&reqs, char*);
    for (size_t i = 0; i < grar_get_size(&reqs); ++i)
    {
        buf[0] = '\0'; // just in case
        if (sprintf(buf, fmt, ids[i]) < 0 ||
            print_func(buf) < 0)
        {
            grar_destroy_with_elements(&reqs, NULL, NULL);
            free(buf);
            return 0;
        }
    }
    
    grar_destroy_with_elements(&reqs, NULL, NULL);
    free(buf);
    return 1;
}

ET2CTestResult
t2c_req_handle_failure(const char* id_str, const char* msg, CReqCat* rc,
    FT2CPrintFunc print_func)
{
    assert(id_str != NULL);
    assert(*id_str != '\0');
    assert(mist_find_first_not_of(id_str, seps, sz_seps) != NULL);
    assert(rc != NULL);
    assert(print_func != NULL);
    
    static const char* str_not_found = "<<< REQ NOT FOUND >>>";
    static const char* str_space = " ";
    static const char* str_one_req = "Requirement failed:";
    static const char* str_many_req = "At least one of the following requirements failed:";
    
    
    if (msg != NULL && *msg != '\0')
    {
        if (print_func(msg) < 0)
        {
            return T2C_TR_ERROR;
        }
    }
    
    CGrowingArray reqs;
    if (!t2c_req_split_list(id_str, &reqs))
    {
        return T2C_TR_ERROR;
    }
    
    size_t num = grar_get_size(&reqs);
    assert(num >= 1);
    
    // output the header
    if (print_func(str_space) < 0 ||
        print_func(num == 1 ? str_one_req : str_many_req) < 0 ||
        print_func(str_space) < 0)
    {
        grar_destroy_with_elements(&reqs, NULL, NULL);
        return T2C_TR_ERROR;
    }
    
    // output the requirements
    static const char* fmt = "{%s}";
    char* buf = (char*)malloc(strlen(fmt) + strlen(id_str) + 1); // should be more than enough
    if (buf == NULL)
    {
        grar_destroy_with_elements(&reqs, NULL, NULL);
        return T2C_TR_ERROR;
    }
    
    int app_req = 0;
    char** ids = grar_get_c_array(&reqs, char*);
    for (size_t i = 0; i < num; ++i)
    {
        if (strstr(ids[i], "app.") == ids[i] ||
            strstr(ids[i], ".app.") != NULL)
        {
            // "app" and non-"app" requirements must not be mixed in one list of IDs
            assert(i == 0 || app_req == 1);
            app_req = 1;
        }
        else
        {
            // "app" and non-"app" requirements must not be mixed in one list of IDs
            assert(app_req == 0);
        }
        
        if (sprintf(buf, fmt, ids[i]) < 0 ||
            print_func(buf) < 0)
        {
            grar_destroy_with_elements(&reqs, NULL, NULL);
            free(buf);
            return T2C_TR_ERROR;
        }
        
        // lookup and output the text of the requirement
        const char* text = t2c_rcat_lookup(rc, ids[i]);
        if (text == NULL)
        {
            text = str_not_found;
        }
        
        if (print_func(text) < 0 ||
            print_func(str_space) < 0)
        {
            grar_destroy_with_elements(&reqs, NULL, NULL);
            free(buf);
            return T2C_TR_ERROR;
        }
    }
    
    grar_destroy_with_elements(&reqs, NULL, NULL);
    free(buf);
    
    if (app_req)
    {
        if (print_func("One or more application requirements failed.") < 0 ||
            print_func("There might be a problem in the test itself. Please report this to the test developer.") < 0)
        {
            return T2C_TR_ERROR;
        }
        return T2C_TR_UNRESOLVED;
    }
        
    return T2C_TR_FAILED;
}

///////////////////////////////////////////////////////////////////////////
// Implementation: private methods
///////////////////////////////////////////////////////////////////////////

int
t2c_req_split_list(const char* id_str, CGrowingArray* reqs)
{
    assert(id_str != NULL);
    assert(reqs != NULL);
    
    if (!grar_create(reqs))
    {
        return 0;
    }
    
    const char* beg = mist_find_first_not_of(id_str, seps, sz_seps);
    const char* end = id_str + strlen(id_str);
    
    if (beg == NULL)
    {
        char* elem = (char*)strdup("");
        if (elem == NULL || !grar_add_element(reqs, elem))
        {
            free(elem);
            grar_destroy(reqs);
            return 0;
        }
        return 1;
    }
    
    while (beg < end)
    {
        const char* pos = mist_find_in_range_first_of(beg, end, seps, sz_seps);
        char* elem = mist_get_substring(beg, pos);
        
        if (elem == NULL || !grar_add_element(reqs, elem))
        {
            free(elem);
            grar_destroy_with_elements(reqs, NULL, NULL);
            return 0;
        }
        
        beg = mist_find_in_range_first_not_of(pos, end, seps, sz_seps);
    }
    
    return 1;
}

int
t2c_req_has_prefixed_req(const char* id_str, const char* prefix)
{
    assert(id_str != NULL);
    assert(prefix != NULL);
    assert(*prefix != '\0');
    assert(strchr(prefix, '.') == NULL);
    
    size_t plen = strlen(prefix);
    assert(plen <= T2C_MAX_PREFIX_LEN);
    
    static char buf[T2C_MAX_PREFIX_LEN + 3]; // +2 for 2 dots max, +1 for '\0'
    buf[0] = '.';
    strcpy(&buf[1], prefix);
    buf[plen + 1] = '.';
    buf[plen + 2] = '\0';
        
    const char* beg = mist_find_first_not_of(id_str, seps, sz_seps);
    const char* end = id_str + strlen(id_str);
    
    if (beg == NULL)
    {
        return 0;
    }
    
    while (beg < end)
    {
        const char* pos = mist_find_in_range_first_of(beg, end, seps, sz_seps);
        
        // [beg, pos) specify a single ID now
        
        // If the ID begins with prefix+'.' or it contains '.'+prefix+'.', 
        // it has the specified prefix.
        if (strstr(beg, &buf[1]) == beg)
        {
            return 1;
        }
        const char* s = strstr(beg, &buf[0]);
        if (s != NULL && s < pos)
        {
            return 1;
        }
        
        beg = mist_find_in_range_first_not_of(pos, end, seps, sz_seps);
    }
        
    return 0;
}
/////////////////////////////////////////////////////////////////////////////
