/// t2c_spawn.c
/// Operations with child processes, alarms, etc.

/////////////////////////////////////////////////////////////////////////////
// Copyright 2009-2010 
// Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//    http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License. 
/////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <assert.h>
    
#include <t2c_spawn.h>

/////////////////////////////////////////////////////////////////////////////
// Constants
/////////////////////////////////////////////////////////////////////////////    

/// Maximum wait time value allowed.
#define T2C_MAX_WAIT_TIME 65535

/////////////////////////////////////////////////////////////////////////////
// Private variables
/////////////////////////////////////////////////////////////////////////////

/// PID of the child process used to execute the test function.
pid_t child_pid = 0;

/// Alarm flag. If zero - no alarm has been raised, nonzero otherwise.
int alarm_raised = 0;

// Format strings for various errors
static const char* errSetTermHandlerFailed = 
    "failed to set handler for signal SIGTERM: %s";
static const char* errSetAlarmFailed = 
    "failed to set alarm: %s";
static const char* errClearAlarmFailed = 
    "failed to clear alarm: %s";
static const char* errChildTimedOut = 
    "child process timed out";
static const char* errWaitFailed = 
    "waiting for child process failed: %s";
static const char* errChildTerminated = 
    "child process was terminated by signal %d";
static const char* errChildStopped = 
    "child process was stopped by signal %d";
static const char* errChildBadStatus = 
    "child process returned bad status (%#x)";


/////////////////////////////////////////////////////////////////////////////
// Declarations: private functions
/////////////////////////////////////////////////////////////////////////////

/// SIGTERM handler. This function performs cleanup: restores default
/// signal actions, kills currently running child process (if any), etc.
void
t2c_sigterm_handler(int sig_num);

/// SIGALRM handler. 
/// The function simply sets a global flag to indicate that the alarm went off.
void
t2c_sigalrm_handler(int sig_num);

/// Set alarm to go off in 'wait_time' seconds. 
/// 'wait_time' must be greater than 0 but not greater than T2C_MAX_WAIT_TIME.
/// The function also sets signal actions as specified by 'act' and returns 
/// previous signal actions in '*old_act'. 'old_act' must not be NULL.
/// The function returns nonzero if successful, 0 otherwise.
int
t2c_set_alarm(unsigned int wait_time, 
    const struct sigaction* act, struct sigaction* old_act);

/// Clear previously set alarm and restore signal actions ('*old_act').
/// The function returns nonzero if successful, 0 otherwise.
/// [NB] The function does not change alarm flag ('alarm_raised').
int
t2c_clear_alarm(const struct sigaction* old_act);

/// Send the specified child process SIGTERM signal, wait a little and then,
/// if the process still exists, kill it with SIGKILL.
/// Sometimes it may take some more time for the child process to finish after 
/// the function returns.
void
t2c_kill_child(pid_t child);

/// This function implements t2c_spawn() if the latter is requested to execute 
/// the test function in the same (caller's) process. See the description of
/// t2c_spawn() for details.
int
t2c_spawn_using_same_process(FT2CTestFunc test_func, 
    FT2CTestStartupFunc startup_func, FT2CTestCleanupFunc cleanup_func,
    ET2CTestResult* test_result,
    char** error_descr);

/// Prepare error message using the specified format (with "%s" for strerror(err_num))
/// If 'error_descr' is not NULL, the error message will be stored in a newly 
/// allocated string, '*error_descr' will point to that string.
/// If 'error_descr' is NULL, the function does nothing.
void
t2c_spawn_format_error_message(char** error_descr, const char* fmt, int err_num);

/// Prepare error message using the specified format using 'num' as an argument.
/// If 'error_descr' is not NULL, the error message will be stored in a newly 
/// allocated string, '*error_descr' will point to that string.
/// If 'error_descr' is NULL, the function does nothing.
void
t2c_spawn_format_error_message_with_num(char** error_descr, const char* fmt, int num);

/////////////////////////////////////////////////////////////////////////////
// Implementation: public functions
/////////////////////////////////////////////////////////////////////////////
int
t2c_spawn(FT2CTestFunc test_func, 
    FT2CTestStartupFunc startup_func, FT2CTestCleanupFunc cleanup_func,
    unsigned int wait_time,
    int single_process,
    ET2CTestResult* test_result,
    char** error_descr)
{
    assert(test_func != NULL);
    assert(test_result != NULL);
    assert(wait_time <= T2C_MAX_WAIT_TIME); // 0 is allowed here
    
    *test_result = T2C_TR_UNRESOLVED; // default test result
    
    if (error_descr != NULL)
    {
        *error_descr = NULL;
    }
    
    // just in case
    fflush(stdout);
    fflush(stderr);
    
    if (single_process)
    {
        // wait_time will be ignored
        return t2c_spawn_using_same_process(test_func, startup_func, cleanup_func,
            test_result, error_descr);
    }

    alarm_raised = 0; // clear the flag even if wait_time == 0    
    child_pid = 0; // default
    pid_t pid = fork();
    switch (pid)
    {
    case -1: // error occured
        if (error_descr != NULL && errno != 0)
        {
            *error_descr = (char*)strdup(strerror(errno));
        }
        return 0;
        
    case 0:  // child process (it should return the result of the test as exit code)
        {
            // reset signal actions to defaults as they have probably been
            // inherited from the parent process
            int ch_sig[] = {SIGTERM, SIGALRM, SIGABRT};
            struct sigaction act;
            
            for (size_t i = 0; i < sizeof(ch_sig)/sizeof(ch_sig[0]); ++i)
            {
                if (sigaction(ch_sig[i], NULL, &act) != -1)
                { 
                    if (act.sa_handler != SIG_IGN && act.sa_handler != SIG_DFL)
                    {
            			act.sa_handler = SIG_DFL;
            		    if (sigaction(ch_sig[i], &act, NULL) == -1)
                        {
                            exit(T2C_TR_UNINITIATED);
                        }
            		}
                }
                else
                {
                    exit(T2C_TR_UNINITIATED);
                }
            } // end for
            
            if (startup_func != NULL)
            {
                if (startup_func() == 0)
                {
                    // initialization failed
                    exit(T2C_TR_UNINITIATED);
                }
            }
            
            ET2CTestResult res = test_func();
            
            if (cleanup_func != NULL)
            {
                cleanup_func();
            }
            
            exit(res);
        } // end of child process's stuff
    default: 
        // parent process (the rest is below this 'switch' statement)
        child_pid = pid;
        break; 
    }
    
    // set handler for SIGTERM to be able to cleanup before exiting
    struct sigaction act;
    struct sigaction old_act;
    if (sigaction(SIGTERM, NULL, &act) != -1)
    {
        if (act.sa_handler == SIG_DFL)
        {
            act.sa_handler = t2c_sigterm_handler;
            if (sigaction(SIGTERM, &act, NULL) == -1)
            {
                t2c_spawn_format_error_message(error_descr, errSetTermHandlerFailed, errno);
                return 0;
            }
        }
    }
    else 
    {
        t2c_spawn_format_error_message(error_descr, errSetTermHandlerFailed, errno);
        return 0;
    }
    
    // if wait_time > 0, set alarm
    if (wait_time > 0)
    {
        act.sa_handler = t2c_sigalrm_handler; 
        act.sa_flags = 0; 
        sigemptyset(&act.sa_mask); // no signals will be blocked
        
        if (t2c_set_alarm(wait_time, &act, &old_act) == 0)
        {
            t2c_spawn_format_error_message(error_descr, errSetAlarmFailed, errno);
            return 0;
        }
    }
    
    // wait until child process finishes or the alarm goes off (if set)
    int status = 0;
    pid_t ret = waitpid(child_pid, &status, WUNTRACED);
    int err = errno; 
    
    if (wait_time > 0)
    {
        if (t2c_clear_alarm(&old_act) == 0)
        {
            t2c_spawn_format_error_message(error_descr, errClearAlarmFailed, errno);
        }
    }
    
    if (ret == -1)  // got a signal or some error has occured
    {
        if (alarm_raised)
        {
            if (error_descr != NULL)
            {
                *error_descr = (char*)strdup(errChildTimedOut);
            }
            *test_result = T2C_TR_UNRESOLVED;
        }
        else
        {
            t2c_spawn_format_error_message(error_descr, errWaitFailed, err);
        }
        
        t2c_kill_child(child_pid);
        return 0;
    }
    
    // if waitpid() returns PID, it must be PID of the particular child process,
    // because there should be only one child
    assert(ret == child_pid); 
    
    if (WIFEXITED(status))
    {
        *test_result = (ET2CTestResult)(WEXITSTATUS(status));
        if (*test_result == T2C_TR_ERROR)
        {
            // The child process has probably called exit with the value of 
            // T2C_TR_ERROR as an argument - map it to T2C_TR_UNRESOLVED because
            // it indicates that there is something wrong with the test rather
            // than with T2C API itself.
            *test_result = T2C_TR_UNRESOLVED;
        }
        return 1; // OK, no errors detected
    }
    else if (WIFSIGNALED(status))
    {
        t2c_spawn_format_error_message_with_num(error_descr, errChildTerminated, (int)(WTERMSIG(status)));
    }
    else if (WIFSTOPPED(status))
    {
        t2c_spawn_format_error_message_with_num(error_descr, errChildStopped, (int)(WSTOPSIG(status)));
        t2c_kill_child(child_pid);
    }
    else
    {
        t2c_spawn_format_error_message_with_num(error_descr, errChildBadStatus, status);
        kill(child_pid, SIGKILL); // just in case
    }
   
    return 0; // some error occured in the child process
}

/////////////////////////////////////////////////////////////////////////////
// Implementation: private functions
/////////////////////////////////////////////////////////////////////////////

void
t2c_sigterm_handler(int sig_num)
{
    assert(sig_num == SIGTERM);
    
    if (child_pid > 0)
    {
        // kill the child if it is still running
        t2c_kill_child(child_pid);
    }
    
    // reset signal handling parameters to the defaults
    struct sigaction act;
    act.sa_flags = 0; 
    act.sa_handler = SIG_DFL;
    sigemptyset(&act.sa_mask); // no signals will be blocked
    sigaction(SIGTERM, &act, (struct sigaction *)NULL);
    
    // send SIGTERM again to itself
    kill(getpid(), SIGTERM);
    
    return;
}

void
t2c_sigalrm_handler(int sig_num)
{
    assert(sig_num == SIGALRM);
    alarm_raised = 1;
    return;    
}

int
t2c_set_alarm(unsigned int wait_time, 
    const struct sigaction* act, struct sigaction* old_act)
{
    assert(wait_time > 0 && wait_time <= T2C_MAX_WAIT_TIME);
    assert(act != NULL);
    assert(old_act != NULL);
    
    alarm(0); // clear any pending alarm 
    if (sigaction(SIGALRM, act, old_act) == -1)
    {
        // failed to set alarm handlers
        return 0;
    }
    
    alarm_raised = 0; // reset the flag
    alarm(wait_time);
            
    return 1;
}

int
t2c_clear_alarm(const struct sigaction* old_act)
{
    assert(old_act != NULL);
    
    alarm(0);
    if (sigaction(SIGALRM, old_act, NULL) == -1)
    {
        return 0; // failed to reset actions to the defaults
    }
    return 1;
}

void
t2c_kill_child(pid_t child)
{
    assert(child > 0);
    
    static const unsigned int wait_time = 10; // wait time in seconds
    
    struct sigaction old_act;
    struct sigaction act;
    act.sa_handler = t2c_sigalrm_handler; 
    act.sa_flags = 0; 
    sigemptyset(&act.sa_mask); // no signals will be blocked
    
    if (kill(child, SIGTERM) == -1) // sending SIGTERM failed
    {
        return; 
    }
    
    // set alarm and wait for the child process to finish
    if (t2c_set_alarm(wait_time, &act, &old_act) == 0)
    {
        return;
    }
    
    // wait for the child to finish but no longer than 'wait_time' seconds
    int status = 0;
    pid_t pid = waitpid(child, &status, 0);
    int err = errno;
    
    if (t2c_clear_alarm(&old_act) == 0)
    {
        return;
    }
    
    if (pid == child)
    {
        // child process has finished, OK
        return;
    }
    if (pid == -1 && !alarm_raised && err != ECHILD)
    {
        errno = err;
        return; // error
    }
    
    if (err == ECHILD) 
    {
        // the specified process is not found, it may have already finished, OK
        return;
    }
    
    // send SIGKILL - just to be sure
    kill(child, SIGKILL);
    return;
}

int
t2c_spawn_using_same_process(FT2CTestFunc test_func, 
    FT2CTestStartupFunc startup_func, FT2CTestCleanupFunc cleanup_func,
    ET2CTestResult* test_result,
    char** error_descr)
{
    assert(test_func != NULL);
    assert(test_result != NULL);
    
    if (startup_func != NULL)
    {
        if (startup_func() == 0)
        {
            // initialization failed
            *test_result = T2C_TR_UNINITIATED;
            return 0;
        }
    }
    
    *test_result = test_func();
    assert(*test_result != T2C_TR_ERROR);
    
    if (cleanup_func != NULL)
    {
        cleanup_func();
    }
    
    return 1; // OK, no errors detected
}

void
t2c_spawn_format_error_message(char** error_descr, const char* fmt, int err_num)
{
    assert(fmt != NULL);
    assert(err_num >= 0);
    
    if (error_descr == NULL)
    {
        return;
    }
    
    char* tstr = strerror(err_num);
    *error_descr = (char*)malloc(strlen(fmt) + strlen(tstr) + 1);
    if (*error_descr != NULL)
    {
        sprintf(*error_descr, fmt, tstr);
    }
    return;
}

void
t2c_spawn_format_error_message_with_num(char** error_descr, const char* fmt, int num)
{
    assert(fmt != NULL);
    
    static const size_t max_digits = 22;
    
    if (error_descr == NULL)
    {
        return;
    }
    
    *error_descr = (char*)malloc(strlen(fmt) + max_digits + 2 + 1); 
    // +2 is for "Ox": %#x can be used in 'fmt'
    
    if (*error_descr != NULL)
    {
        sprintf(*error_descr, fmt, num);
    }
    return;
}
    
/////////////////////////////////////////////////////////////////////////////
