/// t2c_api.c
/// Implementation of T2C API specific to the testing framework used.

/////////////////////////////////////////////////////////////////////////////
// Copyright 2009-2010 
// Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//    http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License. 
/////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>

#include <mist_file_utils.h>
#include <t2c_api.h>

/////////////////////////////////////////////////////////////////////////////
// Private variables
/////////////////////////////////////////////////////////////////////////////

// A FILE* pointer used to output messages to the journal
FILE* flog = NULL;

// Symbolic representations of test results (verdicts)
static const char* res_passed = "PASSED";
static const char* res_unres  = "UNRESOLVED";
static const char* res_failed = "FAILED";
static const char* res_unsupp = "UNSUPPORTED";
static const char* res_untest = "UNTESTED";
static const char* res_uninit = "UNINITIATED";
static const char* res_tmexp  = "TIME_EXPIRED";
static const char* res_fip    = "FIP";
static const char* res_inval  = "INVALID RESULT";

/////////////////////////////////////////////////////////////////////////////
// Implementation: public functions
/////////////////////////////////////////////////////////////////////////////

int 
t2c_printf(const char* format, ...)
{
    assert(format != NULL);
    assert(flog != NULL);
    
    int res = -1;
    va_list ap;
    
    // output spaces (indentation)
    res = fprintf(flog, "    ");
    if (res < 0)
    {
        return res;
    }

    va_start(ap, format);
    res = vfprintf(flog, format, ap);
    va_end(ap);
    
    if (res < 0)
    {
        return res;
    }
    
    res = fprintf(flog, "\n");
    
    // fflush() is necessary: otherwise if the process forks before the data
    // are output, it will be output more than once. Or if the process dies
    // some data may not be written to the log.
    fflush(flog);
    return res;
}

int 
t2c_printf_noindent(const char* format, ...)
{
    assert(format != NULL);
    assert(flog != NULL);
    
    int res = -1;
    va_list ap;
    
    va_start(ap, format);
    res = vfprintf(flog, format, ap);
    va_end(ap);
    
    if (res < 0)
    {
        return res;
    }
    
    res = fprintf(flog, "\n");
    
    // fflush() is necessary: otherwise if the process forks before the data
    // are output, it will be output more than once. Or if the process dies
    // some data may not be written to the log.
    fflush(flog);
    return res;
}

int 
t2c_print_func(const char* arg)
{
    assert(arg != NULL);
    return t2c_printf("%s", arg);
}

void
t2c_print_error(const char* test_name, const char* msg)
{
    assert(test_name != NULL);
    assert(msg != NULL);
    t2c_printf("%s: %s", test_name, msg);
    return;
}    

int
t2c_log_init(const char* suite_root_dir, const char* rel_log_path)
{
    assert(suite_root_dir != NULL);
    assert(rel_log_path != NULL);
    assert(*rel_log_path != '\0');
    assert(flog == NULL); // check for reinitialization
    
    // construct the full path to the log file
    char* plog = t2c_path_sum(suite_root_dir, rel_log_path);
    if (plog == NULL)
    {
        return 1;
    }
    
    // open log
    flog = fopen(plog, "w");
    if (flog == NULL)
    {
        free(plog);
        return 1;
    }
    
    free(plog);
    return 0;
}

void
t2c_log_fini()
{
    assert(flog != NULL);
    fclose(flog);
    flog = NULL;
    
    return;
}

void
t2c_log_flush()
{
    assert(flog != NULL);
    fflush(flog);
    
    return;
}

const char* 
t2c_get_result_string(ET2CTestResult tres)
{
    const char* res = NULL;
    switch (tres)
    {
    case T2C_TR_PASSED:
        res = res_passed;
        break;
    case T2C_TR_UNRESOLVED:
        res = res_unres;
        break;
    case T2C_TR_FAILED:
        res = res_failed;
        break;
    case T2C_TR_UNSUPPORTED:
        res = res_unsupp;
        break;
    case T2C_TR_UNTESTED:
        res = res_untest;
        break;
    case T2C_TR_UNINITIATED:
        res = res_uninit;
        break;
    case T2C_TR_TIME_EXPIRED:
        res = res_tmexp;
        break;
    case T2C_TR_FIP:
        res = res_fip;
        break;
    default:
        res = res_inval;
        break;
    }
    
    return res;
}
/////////////////////////////////////////////////////////////////////////////
