/// t2c_api.h
/// Main header file for T2C API. It includes all other headers required to
/// use T2C API and provides some of the useful macros (e.g. REQ), etc.

/////////////////////////////////////////////////////////////////////////////
// Copyright 2009-2010 
// Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//    http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License. 
/////////////////////////////////////////////////////////////////////////////

#ifndef T2C_API_H_C_TET_1031_INCLUDED
#define T2C_API_H_C_TET_1031_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// These headers provide the part T2C API that does not depend on the testing
// framework (like TET, etc.)
#include <t2c_util.h>
#include <t2c_rcat.h>
#include <t2c_spawn.h>
#include <t2c_req.h>
/////////////////////////////////////////////////////////////////////////////

#ifdef __cplusplus
extern "C" {
#endif

/////////////////////////////////////////////////////////////////////////////
// TRUE and FALSE
/////////////////////////////////////////////////////////////////////////////
#ifndef FALSE
#   define FALSE 0
#   ifndef TRUE
#       define TRUE (!FALSE)
#   endif
#endif

/////////////////////////////////////////////////////////////////////////////
// TRACE
/////////////////////////////////////////////////////////////////////////////

/// Output the message to the journal using the specified format.
/// This is like what printf() does for stdout except that TRACE should be passed
/// at least one argument besides format string 'fmt'.
#define TRACE(fmt, ...) t2c_printf(fmt, __VA_ARGS__)

/// Like TRACE() except that TRACE0() takes no other arguments besides
/// format string 'fmt'.
#define TRACE0(fmt) TRACE("%s", fmt)

/////////////////////////////////////////////////////////////////////////////
// REQ
/////////////////////////////////////////////////////////////////////////////

#define IS_TODO_REQ_EXPR(s_expr) (strstr((s_expr), "TODO_REQ") != NULL)

/// If 'r_expr' does not contain TODO_REQ and the requirements which ids are
/// given in 'r_id' need to be checked, check 'r_expr'.
/// If 'r_expr' expression evaluates to non-zero, simply output that the listed 
/// requirements have been checked. Otherwise output 'r_comment' (if not empty)
/// and then the text of the corresponding requirements via t2c_print_func()
/// and set the result code of the test accordingly.
#define REQ(r_id, r_comment, r_expr) { \
    const char* id_str_ = r_id; \
    if(!IS_TODO_REQ_EXPR(#r_expr) && t2c_req_need_check(id_str_, check_ext_)) {    \
        if(!(r_expr)) { \
            ET2CTestResult tres_ = t2c_req_handle_failure(id_str_, r_comment, rcat_, t2c_print_func); \
            if (tres_ == T2C_TR_ERROR) { \
                t2c_printf("%s, line %d: internal error in REQ", __FILE__, __LINE__); \
                t2c_set_result_code(&test_result_, T2C_TR_UNRESOLVED); \
            } \
            else {t2c_set_result_code(&test_result_, tres_);}  \
            RETURN;                                            \
        } else {                                               \
            if (!t2c_req_print_checked(id_str_, t2c_print_func)) { \
                t2c_printf("%s, line %d: internal error in REQ", __FILE__, __LINE__); \
                t2c_set_result_code(&test_result_, T2C_TR_UNRESOLVED); \
                RETURN; \
            } \
        } \
    } \
}

/// REQva is like REQ except for the following differences.
/// Unlike REQ, the second argument is the expression to be evaluated.
/// The third argument is a printf-style format string for a message to be 
/// output in case of a failure. The format string and the following arguments
/// are passed to TRACE to output this message.
/// That is, REQva takes variable number of arguments.
#define REQva(r_id, r_expr, r_fmt, ...) { \
    const char* id_str_ = r_id; \
    int expr_ = (r_expr); \
    if (!IS_TODO_REQ_EXPR(#r_expr) && t2c_req_need_check(id_str_, check_ext_)) { \
        if (!expr_) { \
            TRACE(r_fmt, __VA_ARGS__); \
        } \
        REQ(r_id, "", expr_);\
    } \
}

/// If TODO_REQ() is used as an expression in REQ or REQva, it indicates that
/// checking of the corresponding feature is yet to be implemented.
#define TODO_REQ() (TRUE)

/////////////////////////////////////////////////////////////////////////////
// Path-related macros
/////////////////////////////////////////////////////////////////////////////

/// Construct the path to the specified test data and returns the resulting
/// string. NULL is returned if there is not enough memory to perform the 
/// operation.
/// The test data directory is <suite_root_dir_>/testdata/<test_name_>.
/// 'rel_path' is a path to the data relative to this directory.
/// The macro does not check if the resulting path or any of its parts exists.
/// The returned pointer should be freed when no longer needed.
#define T2C_GET_DATA_PATH(rel_path) (t2c_get_data_path(suite_root_dir_, test_name_, rel_path))

/// This macro constructs the path to the executable file for the current test
/// and returns the result.
/// The returned pointer should be freed when no longer needed.
#define T2C_GET_TEST_PATH() ((char*)strdup(tet_pname))

/////////////////////////////////////////////////////////////////////////////
// Other macros
/////////////////////////////////////////////////////////////////////////////

/// These constructs mark the beginning and the end of the "finally" section
/// of the test. The contents of the section come from <FINALLY> section 
/// defined in the corresponding <BLOCK> in the t2c-file.
#define BEGIN_FINALLY_SECTION test_finally_section_:
#define END_FINALLY_SECTION

/// RETURN should be used instead of the "return" statement in the test cases.
#define TEST_RETURN return test_result_
#define RETURN                      \
if (in_finally_ == 0)               \
{                                   \
    goto test_finally_section_;     \
}                                   \
else                                \
{                                   \
    TEST_RETURN;                    \
}    

/// This macro should be used in <STARTUP> section to indicate that 
/// the initialization has failed.
/// 'msg' should describe what happened (can be a constant string).
#define INIT_FAILED(msg) {          \
    init_failed_ = TRUE;             \
    reason_to_cancel_ = msg;         \
}

/// This macro aborts test execution and sets the result of the test to 
/// 'FAIL'.
/// Usually one does not need this because test failures are detected by REQ 
/// and REQva macros. 
/// This macro is provided for those rare cases in which using REQ / REQva is
/// inconvenient or even inappropriate. 
/// 'msg' should describe the situation (can be a constant string).
/// [NB] Code in the <FINALLY> section will be executed anyway unless this macro
/// is called from the <FINALLY> section.
#define TEST_FAILED(msg) {    \
    TRACE0(msg);                    \
    t2c_set_result_code(&test_result_, T2C_TR_FAILED);    \
    RETURN;                         \
}

/// This macro should be used in a test to abort its execution if something 
/// wrong happens: a failure to prepare test situation or the means to analyse
/// the resulting state of the system under test, etc.
/// 'msg' should describe the situation (can be a constant string).
/// [NB] Code in the <FINALLY> section will be executed anyway unless this macro
/// is called from the <FINALLY> section.
#define ABORT_TEST_PURPOSE(msg) {   \
    TRACE("Error: %s", msg);        \
    t2c_set_result_code(&test_result_, T2C_TR_UNRESOLVED);     \
    RETURN;                         \
}
#define ABORT_TEST(msg) ABORT_TEST_PURPOSE(msg)

/// This macro should be used in a test to abort its execution if an OPTIONAL 
/// feature to be checked is not supported by the system under test. 
/// 'msg' should describe the situation (can be a constant string),
/// e.g. "Dynamic loading of modules is not supported."
/// [NB] Code in the <FINALLY> section will be executed anyway unless this macro
/// is called from the <FINALLY> section.
#define ABORT_UNSUPPORTED(msg) {    \
    TRACE0(msg);                    \
    t2c_set_result_code(&test_result_, T2C_TR_UNSUPPORTED);    \
    RETURN;                         \
}

/// This macro should be used in a test to abort its execution if a MANDATORY 
/// feature to be checked is not supported by the system under test. 
/// 'msg' should describe the situation (can be a constant string).
/// [NB] Code in the <FINALLY> section will be executed anyway unless this macro
/// is called from the <FINALLY> section.
#define ABORT_UNTESTED(msg) {   \
    TRACE0(msg);                \
    t2c_set_result_code(&test_result_, T2C_TR_UNTESTED);   \
    RETURN;                     \
}

/// This macro should be used in a test to abort its execution and set FIP
/// status ("Further Information Provided"). It indicates that more information
/// is necesary to get about the system under test to decide whether the test
/// passed.
/// 'msg' should describe the situation and, preferably, specify which additional
/// data is necessary ('msg' can be a constant string).
/// [NB] Code in the <FINALLY> section will be executed anyway unless this macro
/// is called from the <FINALLY> section.
#define ABORT_FIP(msg) {   \
    TRACE0(msg);                \
    t2c_set_result_code(&test_result_, T2C_TR_FIP);   \
    RETURN;                     \
}

/////////////////////////////////////////////////////////////////////////////
// TET-specific functions
/////////////////////////////////////////////////////////////////////////////

/// Return TET test result code for the T2C test result code.
/// If T2C_TR_ERROR is passed, behaviour of the function is undefined.
/// The T2C codes for PASSED, FAILED, UNRESOLVED, UNINITIATED, UNTESTED, 
/// UNSUPPORTED are mapped to the corresponding standard TET result codes.
/// T2C_TR_TIME_EXPIRED and T2C_TR_FIP are returned unchanged: these custom
/// result codes should have the same numeric values for TET as for T2C 
/// (check 'tet_code' file to ensure this is the case).
/// All other result codes are mapped to UNRESOLVED. They may occur if the 
/// test calls exit() with the specified value. Nevertheless, it is NOT
/// recommended to call exit() explicitly in the tests anyway.
int
tet_code_for_t2c_code(ET2CTestResult tres);

/// A replacement for tet_printf that takes 'const char*' instead of 'char*'
int 
t2c_printf(const char* format, ...);

/// Same as t2c_printf("%s", arg). 
/// This function should be passed to appropriate t2c_req_* functions, etc.
int 
t2c_print_func(const char* arg);

/// Output the specified error message to the journal. 
/// A prefix consisting of 'test_name' followed by a colon an a space will be 
/// automatically prepended to the message.
/// Use this function to report internal errors only (more often than not, 
/// out of memory conditions or output failures). 
/// Use TRACE macros for all other kinds of output.
void
t2c_print_error(const char* test_name, const char* msg);

/////////////////////////////////////////////////////////////////////////////
#ifdef __cplusplus
}
#endif

#endif // T2C_API_H_C_TET_1031_INCLUDED
