/// parser.h
/// Utilities for parsing .t2c files.

/////////////////////////////////////////////////////////////////////////////
// Copyright 2009-2010 
// Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//    http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License. 
/////////////////////////////////////////////////////////////////////////////

#ifndef PARSER_H_1740_INCLUDED
#define PARSER_H_1740_INCLUDED

#include "smap.h"
#include "mist_string_utils.h"

///////////////////////////////////////////////////////////////////////////  
// Constants
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
/// If a function takes 'beg' and 'end' pointers (usually const char*), it 
/// operates on a string denoted by them. 'beg' points to the beginning of 
/// the string, 'end' - at the first character past the part to be processed.
/// If the whole string is to be processed, 'end' points to the terminating
/// null character.
/// If beg == end, the string is considered empty. Unless specifically stated,
/// this string is allowed to be processed by the function.
/// If 'beg' and / or 'end' are NULL or if they do not point to the same string
/// (end < beg || end > beg + strlen(beg)), 
/// the behaviour of the function is undefined.
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
// Structure definitions
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
/// CT2CParser
/// This structure contains information about the part of the t2c-file that
/// is being parsed.
///////////////////////////////////////////////////////////////////////////
typedef struct CT2CParser_
{
    /// A pointer to the beginning of the string containing the t2c-source.
    const char* source;
    
    /// A pointer to the beginning of the part of the source being parsed.
    const char* beg;

    /// A pointer to the character right past the end of the part of the 
    /// source being parsed.
    const char* end;
    
    /// The value of 'end' saved before matching & parsing a subsection
    const char* old_end;
    
    /// If not NULL, this is a string describing the error encountered
    /// during parsing process.
    char* error;

} CT2CParser;

///////////////////////////////////////////////////////////////////////////
/// EParserResult
/// Possible result codes that can be returned by t2c_parse_* and t2c_match_*
/// functions.
///////////////////////////////////////////////////////////////////////////
typedef enum EParserResult_
{
    /// Success, no error occured in t2c_parse_*.
    PR_PARSE_OK = 0,
    
    /// A syntax  error detected during matching or parsing.
    PR_SYNTAX_ERROR = 1,
    
    /// t2c_match_* function matched the requested structure.
    PR_MATCHED = 2,
    
    /// t2c_match_* function did not match the requested structure.
    PR_NOT_MATCHED = 3,
    
    /// Out of memory
    PR_NO_MEMORY = 4
    
} EParserResult;

///////////////////////////////////////////////////////////////////////////
// Declarations: "public methods"
///////////////////////////////////////////////////////////////////////////

/// Initialize the parser with the specified pointer to the source string, 
/// 'beg' and 'end'. 'error' field will be set to NULL, 'old_end' - to 'end'.
void
t2c_parser_init(CT2CParser* parser, const char* source, 
                const char* beg, const char* end);
                
/// Prepare the parser to process the next chunk of data.
/// 'beg' will be set to the value of 'end', 'end' - to 'old_end'.
/// Call this function only if the previous section has been matched
/// successfully.
void
t2c_parser_advance(CT2CParser* parser);

/// Return nonzero if the parser structure contains valid data, 0 otherwise.
/// This function will be used the most in assertions.
int
t2c_is_parser_valid(CT2CParser* parser);

/// If error != NULL, *error_descr = parser->error, otherwise
/// just free 'parser->error' if necessary and do nothing more.
/// No deep copy happens here.
void
t2c_set_error(char** error_descr, CT2CParser* parser);

/// Initializes parser->error with a newly allocated string that has the 
/// following format:
/// line %u: <fmt_with_arg_substituted_for_%s>
/// If 'arg' is NULL, 'fmt' may not contain "%s".
/// Line number is determined from parser->beg and parser->source.
void 
t2c_format_parse_error(CT2CParser* parser, const char* fmt, const char* arg);

/// Like t2c_format_parse_error() except that this function does not output line
/// number information
void 
t2c_format_parse_error_no_line(CT2CParser* parser, const char* fmt, const char* arg);

///////////////////////////////////////////////////////////////////////////
/// t2c_parse_* and t2c_match_* functions
///
/// Each of these functions operates on the part of the source defined by 
/// [parser->beg, parser->end), referred to as [beg, end) below, for brevity.
///
/// If a t2c_match_* function matches the requested structure, it adjusts
/// 'beg' and 'end' so that they correspond to the contents of the structure.
/// For example, if a function matches a <BLOCK>...</BLOCK> section, it will
/// set 'beg' to point right past "<BLOCK>" and 'end' - at '<' character in
/// "</BLOCK>" tag.
/// If the function does not match the requested structure or an error occurs,
/// 'beg' and 'end' remain unchanged.
///
/// In case of failure, t2c_parse_* and t2c_match_* functions may return 
/// the description of the error in parser->error. This string must be 
/// freed when no longer needed.
/// 
/// On success, t2c_parse_*_section functions set parser->end to point to the
/// character just after '>' of the closing tag. This facilitates the use of
/// t2c_parser_advance() after t2c_parse_*_section().
///////////////////////////////////////////////////////////////////////////

/// Match a section.
/// 'name' is the name of the section to be matched.
/// If matched the section, the function sets 'beg' and 'end' so that
/// they define the contents of the section. In addition, if 'attr' is not NULL,
/// string map '*attr' will be initialized by the function and the attributes 
/// of the section (i.e. of its opening tag) will be stored there.
/// '*attr' will be set to NULL in case of failure and also if there are no 
/// attributes defined for this section.
EParserResult
t2c_match_section(CT2CParser* parser, const char* name, CStringMap** attr);

/// Parse a section that is expected to contain no subsections.
/// (e.g. GLOBAL, STARTUP, CLEANUP; CODE and FINALLY subsections in
/// BLOCK).
/// The function just sets the contents of the section as a value of 
/// the parameter in 'params' with the name specified in 'name'.
EParserResult
t2c_parse_simple_section(CT2CParser* parser, const char* name, CStringMap* params);

/// Match opening tag with specified name. 
/// Each opening tag has the following form:
///     <NAME [attr0="val0" [attr1="val1" ...]]>
/// The search is case sensitive. Spaces and tabs are ignored before and after
/// the name of the tag and names and (quoted) values of the attributes (if any).
/// Each value of an attribute should be enclosed in double quotes (").
///
/// Whitespace characters are not allowed within the name of the tag.
///
/// If there are attributes in the tag and 
/// attr != NULL, create '*attr' string map and populate it with the attributes.
/// If attr != NULL but there are no attributes in this tag, set '*attr' to NULL.
/// Multiple attributes with the same name are not allowed.
///
/// If matched, this function sets parser->beg to point past the '>' character.
EParserResult
t2c_match_opening_tag(CT2CParser* parser, const char* tag_name, CStringMap** attr);

/// Try to find closing tag with specified name. 
/// Each closing tag has the following form:
///     </NAME>
/// The search is case sensitive. Spaces and tabs are ignored before and after
/// the name of the tag.
/// 
/// If matched, this function sets parser->end to point to the '<' character
/// of the closing tag.
EParserResult
t2c_match_closing_tag(CT2CParser* parser, const char* tag_name);

/// Parse the specified substring [beg, end), create '*attr' string map and 
/// populate '*attr' with respective {attribute_name, attribute_value} pairs.
/// See the description of t2c_match_opening_tag() for the required format
/// of the substring and for other restrictions.
EParserResult
t2c_parse_attributes(CT2CParser* parser, CStringMap** attr);

/// Set parser->end to point to the character just after the first '>' of parser->end.
/// Can be used to position parser->end to point after the closing tag.
void
t2c_parser_to_section_end(CT2CParser* parser);


#endif // PARSER_H_1740_INCLUDED
