/// t2c.c
/// T2C test code generator.

/////////////////////////////////////////////////////////////////////////////
// Copyright 2009-2010 
// Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//    http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License. 
/////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <getopt.h>
#include <assert.h>

#include <limits.h>
#include <dirent.h>
#include <time.h>
#include <errno.h>

#include "about.h"
#include "mist_base.h"
#include "mist_file_utils.h"
#include "mist_string_utils.h"
#include "t2c.h"
#include "parser.h"

///////////////////////////////////////////////////////////////////////////  
// Constants
///////////////////////////////////////////////////////////////////////////

/// Name of the parameter defined in the .cfg file of the suite, that specifies
/// the target collection of templates to use.
#define PNAME_TARGET        "TARGET"

/// Name of the parameter defined in the .cfg file of the suite, that specifies
/// the additional template directory. The parameter can be used to override 
/// some of the templates from the main template directory or add new ones.
#define PNAME_TEMPLATE_DIR  "TEMPLATE_DIR"

/// Name of the test case template
#define TC_TPL_NAME "test_case"

// T2C directives
#define T2C_DIR_LIBRARY     "library"
#define T2C_DIR_LIBSECTION  "libsection"
#define T2C_DIR_ADD_RCAT    "additional_req_catalogues"

///////////////////////////////////////////////////////////////////////////
// Structure definitions
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
/// EVRElementType
/// Type of an element in a value record
///////////////////////////////////////////////////////////////////////////
enum EVRElementType
{
    VRE_STRING,
    VRE_RANGE
};

///////////////////////////////////////////////////////////////////////////
/// CValueElement - an element of a value record in a VALUES / PURPOSE
/// section.
///////////////////////////////////////////////////////////////////////////
typedef struct CValueElement_
{
    /// Type of the element
    enum EVRElementType type; 
    
    /// the contents of the element of type VRE_STRING
    char* str;
    
    /// Minimum and maximum values of the element of type VRE_RANGE
    /// (SET-constructs only)
    long vmin;
    long vmax;
    
    /// Number of times to repeat the element (RES-constructs only)
    long ntimes;
    
} CValueElement;

///////////////////////////////////////////////////////////////////////////
/// EVRecordType
/// Type of a value record
///////////////////////////////////////////////////////////////////////////
enum EVRecordType
{
    VR_PLAIN,   /// neither SET-construct nor RES-construct
    VR_SET,     /// SET(...)
    VR_RES      /// RES(...)
};

///////////////////////////////////////////////////////////////////////////
/// CValueRecord - a value record in a VALUES / PURPOSE section.
///////////////////////////////////////////////////////////////////////////
typedef struct CValueRecord_
{
    /// Type of the record
    enum EVRecordType type;
    
    /// Elements defined by the record. Plain records have only one,
    /// SET- or RES- constructs may have more.
    CGrowingArray elem;
} CValueRecord;

///////////////////////////////////////////////////////////////////////////
// Private variables
///////////////////////////////////////////////////////////////////////////

// Section names
static const char* sec_global   = "GLOBAL";
static const char* sec_startup  = "STARTUP";
static const char* sec_cleanup  = "CLEANUP";
static const char* sec_block    = "BLOCK";
static const char* sec_targets  = "TARGETS";
static const char* sec_define   = "DEFINE";
static const char* sec_code     = "CODE";
static const char* sec_finally  = "FINALLY";
static const char* sec_purpose  = "PURPOSE";
static const char* sec_values   = "VALUES";

// Whitespace characters, etc.
static const char wspace[] = " \t\n\r";
static size_t sz_wspace = sizeof(wspace) / sizeof(wspace[0]) - 1;

static const char ws[]  = " \t";
static size_t sz_ws = sizeof(ws) / sizeof(ws[0]) - 1;

static const char eol[] = "\n\r";
static size_t sz_eol = sizeof(eol) / sizeof(eol[0]) - 1;

static const char ws_and_digits[] = " \t-0123456789";
static size_t sz_ws_and_digits = sizeof(ws_and_digits) / sizeof(ws_and_digits[0]) - 1;

// Format strings for various errors
static const char* errBadLineInHeader = 
    "T2C header may contain only T2C directives or comments";
/*static const char* errUnknownDirInHeader = 
    "unknown directive in T2C header";*/
static const char* errParamAlreadySet = 
    "parameter '%s' is already set";
static const char* errDirInvalid = 
    "invalid '#%s' directive";
static const char* errNoSectionsFound = 
    "no sections found in the .t2c file";
static const char* errUnexpectedClosingTag = 
    "found unexpected closing tag";
static const char* errBlockExpected = 
    "<BLOCK> section expected";
static const char* errTargetsExpected = 
    "<TARGETS> section expected";
static const char* errCodeExpected = 
    "<CODE> section expected";
static const char* errUnexpectedTextInBlock = 
    "unexpected text found in <BLOCK> section (please check if </BLOCK> tag is in place)";
static const char* errNoTargets = 
    "no targets specified in <TARGETS> section";
static const char* errBadDefine = 
    "invalid record in <DEFINE> section (expected: \"#define <NAME> [VALUE]\")";
static const char* errTCMultivalued = 
    "evaluation of test case template gave multivalued result; please ensure "
    "you use join() for all multi-valued entities it refers to";
static const char* errTplGenFail = 
    "failed to generate template for the tests from the block #%d: %s";
static const char* errTplGenFailNoMem = 
    "failed to generate template for the tests from the block #%d: not enough memory";
static const char* errDiffNumVals = 
    "values are specified for different number of parameters in the <VALUES> and/or <PURPOSE> sections";
static const char* errMissingLeftParen = 
    "missing '(' in the %s-construct";
static const char* errMissingRightParen = 
    "missing ')' in the %s-construct";
static const char* errEmptyElement = 
    "found an empty element in the %s-construct";
static const char* errLoneQuot = 
    "found lone quotation mark (\")";
static const char* errLoneApos = 
    "found lone apostrophe (\')";
static const char* errInvalidSETElement = 
    "invalid element in the SET-construct: \'%s\'";
static const char* errInvalidRESElement = 
    "invalid element in the RES-construct: \'%s\'";
static const char* errOnlyPosAllowed = 
    "only positive numbers are allowed after ':' in a RES-construct";
static const char* errConvertToNumFailed = 
    "\'%s\' is not a number or it is out of the allowed range";
static const char* errUnexpectedAttribs = 
    "unexpected attributes in <%s> section";
static const char* errEmptyFile = 
    "the file is empty: \"%s\"";
static const char* errNoDataDir = 
    "data directory \"%s\" does not exist";
static const char* errDataDirIsRoot = 
    "root directory should not be used as a data directory";
static const char* errFailedToConstructDataPath = 
    "unable to construct path to the data directory";
static const char* errNoTemplateDir = 
    "template directory \"%s\" does not exist";
static const char* errNoSuiteDir = 
    "test suite directory \"%s\" does not exist";
static const char* errNoFile = 
    "file \"%s\" does not exist or cannot be accessed";
static const char* errFailedToLoadCfg = 
    "failed to load configuration for the suite from \"%s\"";
static const char* errFailedToLoadCfgWithErr = 
    "failed to load configuration for the suite from \"%s\": %s";
static const char* errFailedToLoadTGCfg = 
    "failed to load configuration for the test group \"%s\" from \"%s\"";
static const char* errFailedToLoadTGCfgWithErr = 
    "failed to load configuration for the test group \"%s\" from \"%s\": %s";
static const char* errDupParam = 
    "parameter \"%s\" is specified more than once in the .cfg file for the suite";
static const char* errDupParamTG = 
    "parameter \"%s\" is specified more than once in the .cfg file "
    "for the test group \"%s\"";
static const char* errInvalidParamTG = 
    "it is not allowed to specify \"%s\" parameter in the .cfg file "
    "for the test group \"%s\"";
static const char* errMissingTarget = 
    "please specify the target to generate files for "
    "(either using -t option or in \"%s\" parameter in the .cfg file of the suite)";
static const char* errInvalidTargetName = 
    "invalid name of a target: \"%s\"";
static const char* errNoAuxTemplateDir = 
    "template directory specified for the suite does not exist: \"%s\"";
static const char* errFailedToSetStdParams = 
    "failed to set standard parameters for the tests";
static const char* errDirReadFailed = 
    "unable to read the contents of \"%s\"";
static const char* errReadTplListFailed = 
    "unable to read the list of templates in \"%s\"";
static const char* errNoTemplates = 
    "there are no templates to instantiate for target \"%s\"";
static const char* errValueAlreadySpecified = 
    "\"%s\" is already specified in the configuration file(s) for the test group";
static const char* errGenTestsFail = 
    "failed to generate tests for \"%s\"";
static const char* errTplLoadFail = 
    "failed to load templates from \"%s\"";
static const char* errOutFileConstructFail = 
    "failed to construct the string containing path to the output file";
static const char* errFailedToSetTestParams = 
    "failed to set test-specific parameters for the tests generated from \"%s\"";
static const char* errGenOutputFail = 
    "failed to generate output files for the templates from \"%s\"";
static const char* errT2CFileLoadFailed = 
    "failed to load .t2c file: \"%s\"";
static const char* errTGNotExist = 
    "template group \"%s\" does not exist";
static const char* errTemplateDirEmpty = 
    "TEMPLATE_DIR parameter is defined but has empty value";

///////////////////////////////////////////////////////////////////////////  
// Declarations of "private" functions
///////////////////////////////////////////////////////////////////////////

/// Output the specified error message.
void
error_message(const char* msg);

/// Output the error message corresponding to the error code.
void 
error_message_for_mist_code(EMistErrorCode ec);

/// Process the arguments specified in the command line.
/// See the description of init() function for details.
int
load_args(int argc, char* argv[], struct SSettings* psettings);

/// Show usage information.
void
show_usage();

/// Show version information.
void
show_version();

/// Return the path to the T2C data directory. 
/// If $T2C_DATADIR environment variable is defined and is not empty, it is used.
/// Otherwise DATADIR is used (specified in -D compiler option).
/// The returned string should be freed when no longer needed.
/// In case the process is out of memory, NULL is returned.
/// This function does not check if the directory actually exists. 
char*
get_data_dir();

/// Output the path to the T2C data directory to stdout (see get_data_dir()).
/// Return nonzero if successful, 0 otherwise (if there is not enough memory).
int
print_data_dir();

/// A helper function that constructs the path to the per-test template 
/// based on the path to the .t2c file (note that the file may not have t2c extension).
/// The template directory is assumed to be in the same directory as the file
/// and to have the same name (if the file has t2c extension, the latter is 
/// not copied to the resulting name, other extensions are copied). 
/// ".templates" extension is then appended to the resulting name.
/// 
/// Examples:
/// template_dir_for_file_path("tests/foo.t2c") => "tests/foo.templates"
/// template_dir_for_file_path("/home/tests/bazbar") => "/home/tests/bazbar.templates"
/// template_dir_for_file_path("/home/tests/foo.bar") => "/home/tests/foo.bar.templates"
/// 
/// The function does not check if the template directory exists, it just
/// constructs the path to it.
char* 
template_dir_for_file_path(const char* file_path);

/// Read the contents of the specified directory and update the string map 
/// accordingly. That is, for each subdirectory of the directory, 'smap' should
/// contain an element with the name of the subdir as a key and the full path 
/// to it as a value.
/// If there is an element for this subdirectory name, the corresponding value is 
/// updated, otherwise a new element is created. Other elements remain unchanged.
/// 'smap' must be initialized before calling this function.
/// Usually you should pass an absolute path here rather than a relative one.
/// If the specified directory does not exists, the function does nothing. 
/// It is not considered an error.
/// The function returns nonzero if successful, 0 otherwise.
int
update_template_list(const char* dir_path, struct CStringMap_* smap);

/// Add standard parameters with appropriate values to the collection of parameters:
/// - T2C_DATADIR - path to the t2c data directory (this directory contains
///   'templates' subdirectory, etc.)
/// - SUITE_ROOT_DIR is the absolute path to the root directory of the suite;
/// - SUITE_NAME is the name of the suite (without "-t2c" suffix if it was present);
/// - T2C_FILE_NAME is the name of the .t2c file (without "t2c" extension if it 
///   was present) - set in per-test mode only;
/// - YEAR is the current year in YYYY format (mostly for copyright notices).
/// - TEST_MODULE - the collection of records (names or paths of the tests) 
///   listed in the file specified in '--list-file' option.
///
/// The function returns nonzero if successful, 0 otherwise.
int 
set_std_params(struct SSettings* psettings);

/// If the file specified in '--list-file' option exists and contains something
/// except separator characters (CR, LF, TAB or space), the records are read 
/// from this file to become values of 'TEST_MODULE' attribute.
/// 
/// The function returns nonzero if successful, 0 otherwise (basically, if 
/// there is not enough memory to complete the operation).
int 
populate_test_module_list(struct SSettings* psettings);

/// Load .t2c file, process the specified template group for the test cases 
/// ('tc_group_name'), prepare a string map with test-specific parameters 
/// ('params').
/// The path to the .t2c file is in '*psettings' structure as well as the common
/// parameters (they are used in evaluation of the test case template group).
/// The function returns a pointer to the newly created string map with 
/// test-specific parameters successful, NULL in case of failure.
CStringMap*
process_t2c_file(struct SSettings* psettings, const char* tc_group_name);

/// Load the group of templates for test cases (named 'tc_group_name').
/// The function returns a pointer to the template group on success, NULL on
/// failure.
CMistTemplateGroup*
load_tc_template_group(struct SSettings* psettings, const char* tc_group_name);

/// Parse the t2c source ('source') read from a .t2c file and load data from it,
/// store global attributes in 'params'. 
/// 'tc_group' template group is used to generate the test cases.
/// 
/// The function returns nonzero if successful, 0 otherwise.
///
/// If the function fails and error_descr != NULL, the description of the error
/// may be returned in *error_descr. This string should be freed by the caller
/// when no longer needed. Note that it is not mandatory for the function to 
/// return anything in *error_descr.
int
t2c_load_data(const char* source, CStringMap* params, 
               CMistTemplateGroup* tc_group, char** error_descr);
               
/// Read the .t2c file specified by 'path' skipping #-comments and return a
/// pointer to the resulting string. The string should be freed by the called 
/// when no longer needed.
/// NULL is returned in case of failure.
char*
t2c_read_file(const char* path);

/// Return nonzero if the string specified by ['beg', 'end') begins with a name
/// of C/C# preprocessor directive (according to C99 standard (ISO/IEC 9899:1999)
/// and C# Language Reference - MSDN) or a T2C directive, 0 otherwise.
int
t2c_is_directive(const char* beg, const char* end);

/// Return nonzero if 'str' is a t2c-style line comment, i.e. it begins with 
/// '#' and is not a directive, 0 otherwise.
/// A directive begins with '#' (optionally followed by space or '\t') followed
/// by the name of the directive.
/// Space and '\t' characters are allowed before '#'.
int 
t2c_is_comment(const char* beg, const char* end);

///////////////////////////////////////////////////////////////////////////
/// t2c_parse_* and t2c_match_* functions
///
/// Each of these functions operates on the part of the source defined by 
/// [parser->beg, parser->end), referred to as [beg, end) below, for brevity.
///
/// If a t2c_match_* function matches the requested structure, it adjusts
/// 'beg' and 'end' so that they correspond to the contents of the structure.
/// For example, if a function matches a <BLOCK>...</BLOCK> section, it will
/// set 'beg' to point right past "<BLOCK>" and 'end' - at '<' character in
/// "</BLOCK>" tag.
/// If the function does not match the requested structure or an error occurs,
/// 'beg' and 'end' remain unchanged.
///
/// In case of failure, t2c_parse_* and t2c_match_* functions may return 
/// the description of the error in parser->error. This string must be 
/// freed when no longer needed.
/// 
/// On success, t2c_parse_*_section functions set parser->end to point to the
/// character just after '>' of the closing tag. This facilitates the use of
/// t2c_parser_advance() after t2c_parse_*_section().
///////////////////////////////////////////////////////////////////////////

/// Match T2C header.
EParserResult
t2c_match_header(CT2CParser* parser);

/// Parse T2C header, extract the parameters specified in T2C directives.
EParserResult
t2c_parse_header(CT2CParser* parser, CStringMap* params);

/// Parse <BLOCK> section. 
/// Create a value of "test_case" parameter for each test case defined in the 
/// block. 
/// 'tc_group' is used to generate the source of each test case.
/// 'attr' is the string map filled with the block's attributes by t2c_parse_block_section(). 
/// NULL if the block has no attributes.
/// 'block_index' is an index of the block in the .t2c file (starting with 1).
/// '*abs_test_index' is incremented for each test in the block. This is used 
/// to provide <$abs_test_index$> parameter.
EParserResult
t2c_parse_block_section(CT2CParser* parser, int block_index, int* abs_test_index,
    CStringMap* params, CMistTemplateGroup* tc_group, CStringMap* attr);

/// Parse '#library' directive. Set its contents as a value of the 
/// respective parameter. Offset 'parser->beg' to point to the next line.
EParserResult
t2c_parse_library_directive(CT2CParser* parser, CStringMap* params);

/// Parse '#libsection' directive. Set its contents as a value of the 
/// respective parameter. Offset 'parser->beg' to point to the next line.
EParserResult
t2c_parse_libsection_directive(CT2CParser* parser, CStringMap* params);

/// Parse '#additional_req_catalogues' directive. Set its contents as values
/// of the respective parameter. Offset 'parser->beg' to point to the next line.
EParserResult
t2c_parse_add_rcat_directive(CT2CParser* parser, CStringMap* params);

// [NB] <CODE> and <FINALLY> sections can be parsed using t2c_parse_simple_section(),
// no special functions are needed for this.

/// Parse <TARGETS> section. 
/// Each non-blank line of this section is considered a target of the test
/// (leading and trailing whitespace are ignored).
/// The section must define at least one target.
/// The functions sets each target as a value of "target" parameter (it can 
/// be therefore multivalued).
EParserResult
t2c_parse_targets_section(CT2CParser* block_parser, CStringMap* block_params);

/// Parse <DEFINE> section. 
/// Each non-blank line in this section must have the following format:
///  #define <DNAME> [<DVALUE>]
/// The section may be empty (i.e. contain no non-blank lines)
/// <DNAME> and <DVALUE> values will be set as the values of "define_name" and
/// "define_value" parameters, respectively. 
/// If <DVALUE> is omitted, it is considered an empty string.
EParserResult
t2c_parse_define_section(CT2CParser* block_parser, CStringMap* block_params);

/// Parse <VALUES> (<PURPOSE>) section and generate the complete source of the
/// test case using 'tg' template group. 
/// Set the value of "test_case" parameter of the test group as well as 
/// "block_index", "test_index" and "abs_test_index" global parameters.
/// '*test_index' and '*abs_test_index' will be updated accordingly.
/// 'attr' contains the attributes of the block (can be NULL).
/// The function will return number of (sets of) values in the section in
/// '*num_values'. This can be used to check if each <VALUES> section in the 
/// block have the same number of (sets of) values specified.
EParserResult
t2c_parse_values_section(CT2CParser* block_parser, 
    int block_index, int* test_index, int* abs_test_index,
    int* num_values,
    CMistTemplateGroup* tg,
    CStringMap* params, CStringMap* attr);

/// Process all <VALUES> (<PURPOSE>) sections (see t2c_parse_values_section)
EParserResult
t2c_process_values_sections(CT2CParser* block_parser, 
    int block_index, int* abs_test_index,
    CMistTemplateGroup* tg,
    CStringMap* params, CStringMap* attr);
    
/// Generate the sources of the test cases for the parameter value records
/// specified in 'records'. 'tg' template group is used to do this.
/// The function updates the test index in the block ('*test_index') and the 
/// absolute test index ('*abs_test_index') as it processes the records 
/// and generates tests.
/// Each generated source is stored as a value of "test_case" parameter in 
/// 'params' string map.
/// The function returns nonzero if successful, 0 if there is not enough memory
/// to perform the operation.
int
t2c_parser_gen_test_sources(CGrowingArray* records, 
    int block_index, int* test_index, int* abs_test_index,
    CMistTemplateGroup* tg,
    CStringMap* params, CStringMap* attr);

/// Set the values of special local and global parameters ('*_index') as well as 
/// the local parameters specified in 'local_params' and 'attr' (if attr != NULL) 
/// and generate the source of the test.
/// The resulting source is set as a value of "test_case" global parameter.
/// The function returns nonzero if successful, 0 if there is not enough memory 
/// to perform the operation.
int
t2c_generate_test_source(int block_index, int test_index, int abs_test_index,
    CMistTemplateGroup* tg,
    CStringMap* local_params, CStringMap* global_params, CStringMap* attr);
    
/// Perform a depth-traversal of 'vr' array containing 'nvals' elements starting
/// from the element with index 'start'. Prepare the corresponding set of values 
/// and generate the sources of the tests (set them as values of "test_case" global
/// parameter).
/// 'pvals' array should contain at least 'nvals' elements, it will be used by the 
/// function for temporary storage of the values.
/// The function returns nonzero if successful, 0 otherwise.
int
t2c_traverse_value_list(CValueRecord** vr, const char** pvals, size_t start, size_t nvals,
    int block_index, int* test_index, int* abs_test_index,
    CMistTemplateGroup* tg, CStringMap* global_params, CStringMap* attr);
    
/// Set the values of numbered parameters (<%0%>, <%1%>, ...) and their collection
/// (multivaled parameter <%param%>) in 'local_params'.
/// The values are listed in 'pvals', there are 'nvals' of them.
/// The function returns nonzero if successful, 0 otherwise.
int 
t2c_set_numbered_params(const char** pvals, size_t nvals, CStringMap* local_params);

/// Find the element of the RES-construct corresponding to the test case number 
/// 'tc_num'. The latter is a number of the test case among those defined by 
/// currently processed <VALUES> / <PURPOSE> section.
/// The function should always returns a valid pointer.
CValueElement*
t2c_find_res_element(CValueRecord* vr, long tc_num);

/// Set the values "block_index", "test_index" and "abs_test_index" global 
/// parameters.
/// The function returns nonzero if successful, 0 otherwise.
int
t2c_set_index_values(CStringMap* params, 
    int block_index, int test_index, int abs_test_index);
    
/// Create a template group for the block number 'block_index' to generate test
/// case sources from.
/// First, the function substitutes parameters from 'block_params' to 'tc_group'
/// and evaluates 'tc_group'. The result should be single-valued.
/// Then a new template group is loaded from the resulting string with 
/// "<%" and "%>" as placeholder markers. 
/// The function returns the pointer to the newly created template group, or
/// NULL in case of failure.
/// In the latter case, the function may also return a description of the failure
/// in '*error_descr' if 'error_descr' is not NULL. '*error_descr' should be
/// freed when no longer needed.
CMistTemplateGroup* 
t2c_prepare_test_tg(CMistTemplateGroup* tc_group, int block_index, 
    CStringMap* block_params, char** error_descr);

///////////////////////////////////////////////////////////////////////////
// Value record parsing utilities
///////////////////////////////////////////////////////////////////////////
    
/// Create CValueElement structure (type: string). 
/// The contents are copied from [beg; end), 'ntimes' field is set to 'ntimes'.
/// Return a pointer to the newly created structure if successful, NULL otherwise.
CValueElement*
t2c_vr_create_string_element(const char* beg, const char* end, long ntimes);

/// Create CValueElement structure (type: range) with the fields specified.
/// If vmin > vmax, these values will be reversed.
/// Return a pointer to the newly created structure if successful, NULL otherwise.
CValueElement*
t2c_vr_create_range_element(long vmin, long vmax);

/// A destructor for CValueElement structures.
/// Can be handy for grar_destroy_with_elements().
void 
t2c_vr_element_dtor(void* doomed, void* user_data);

/// Create the CValueRecord structure of the specified type with no elements.
/// Return a pointer to the newly created structure if successful, NULL otherwise.
CValueRecord*
t2c_vr_create_value_record(enum EVRecordType type);

/// A destructor for CValueElement structures.
/// Can be handy for grar_destroy_with_elements().
void 
t2c_vr_value_record_dtor(void* doomed, void* user_data);

/// Parse the specified string to produce a value record. The string may define
/// a plain value, a SET-construct or a RES-construct.
/// The pointer to the resulting record structure is returned in *rec on success.
/// In case of a failure, the function may return the description of the error
/// in *error_descr. This string should be freed when no longer needed.
/// 'error_descr' must point to a valid pointer when the function is called.
/// The function will not change the contents of 'rstr'.
EParserResult
t2c_vr_parse_value_record(CMistString* rstr, CValueRecord** rec, char** error_descr);

/// Match <name>(...) construct (SET(), RES(), ...)
EParserResult
t2c_vr_match_construct(CT2CParser* vr_parser, const char* name);

/// Parse the contents of a SET(...) or RES(...) construct, save the extracted
/// elements in '*rec'.
/// The type of the construct is determined from rec->type.
EParserResult
t2c_vr_parse_construct(CT2CParser* vr_parser, CValueRecord* rec);

/// Parse the element of a SET(...) construct, save the extracted data
/// in '*rec'.
EParserResult
t2c_vr_parse_set_element(CT2CParser* vr_parser, CValueRecord* rec);

/// Match a range in a SET-construct:
/// <INTEGER>..<INTEGER>
/// If the element is matched, the function parses it as well and adds appropriate
/// element to 'rec'.
EParserResult
t2c_vr_match_range_element(CT2CParser* vr_parser, CValueRecord* rec);

/// Parse the element of a RES(...) construct, save the extracted data
/// in '*rec'.
EParserResult
t2c_vr_parse_res_element(CT2CParser* vr_parser, CValueRecord* rec);

/// Save the plain record in '*rec'.
EParserResult
t2c_vr_process_plain_record(CT2CParser* vr_parser, CValueRecord* rec);


///////////////////////////////////////////////////////////////////////////  
// Implementation
///////////////////////////////////////////////////////////////////////////

void
show_usage()
{
    printf("%s\n%s\n", T2C_LONG_NAME, T2C_USAGE);
    return;
}

void
show_version()
{
    printf("%s version %s\n", T2C_SHORT_NAME, T2C_VERSION);
    return;
}

int
load_args(int argc, char* argv[], struct SSettings* psettings)
{
    int c;
    int ret = 1;

    assert(psettings != NULL);

    struct option long_options[] =
    {
        // these options do not set flags
        {"help",    no_argument, NULL, 'h'},
        {"version", no_argument, NULL, 'v'},
        {"templates",   required_argument, NULL, 'd'},
        {"target",      required_argument, NULL, 't'},
        {"list-file",   required_argument, NULL, 'f'},
        {"print-data-dir", no_argument, NULL, 'p'},
        {NULL, 0, NULL, 0}
    };
    
    while (1)
    {
        int option_index = 0;
        c = getopt_long(argc, argv, "t:d:f:", long_options, &option_index);
        
        if (c == -1) // all options have been processed
        {
            break;
        }
        
        switch (c)
        {
        case 0:
            // If a flag was set, nothing more to do. If not...
            if (long_options[option_index].flag == 0)
            {
                assert(0); // should not get here
            }
            break;
            
        case 'h':
            show_usage();
            
            cleanup_settings(psettings);
            exit(EXIT_SUCCESS);
            break;
            
        case 'v':
            show_version();
            
            cleanup_settings(psettings);
            exit(EXIT_SUCCESS);
            break;
            
        case 'd':
            if (optarg[0] == '\0')
            {
                error_message("it is not allowed to specify an empty path to the template directory");
                return 0;
            }
            
            psettings->template_dir = mist_path_absolute(optarg);
            if (psettings->template_dir == NULL)
            {
                
                error_message("unable to obtain the absolute path to the template directory");
                return 0;
            }
            break;
            
        case 't':
            if (optarg[0] == '\0')
            {
                error_message("it is not allowed to specify an empty target");
                return 0;
            }
            
            psettings->target = (char*)strdup(optarg);
            if (psettings->target == NULL)
            {
                error_message_for_mist_code(MIST_OUT_OF_MEMORY);
                return 0;
            }
            break;
        
        case 'f':
            if (optarg[0] == '\0')
            {
                error_message("it is not allowed to specify an empty path to the test list file");
                return 0;
            }
        
            psettings->test_name_file_path = mist_path_absolute(optarg);
            if (psettings->test_name_file_path == NULL)
            {
                error_message("unable to obtain the absolute path to the test list file");
                return 0;
            }
            break;
            
        case 'p':
            ret = print_data_dir();
            cleanup_settings(psettings);
            
            if (ret)
            {
                exit(EXIT_SUCCESS);
            }
            else
            {
                error_message_for_mist_code(MIST_OUT_OF_MEMORY);
                exit(EXIT_FAILURE);
            }
            
            break;

        case '?':
            // getopt_long should have already printed an error message.
            ret = 0;
            break;

        default:
            assert(0); // should not get here
            break;
        }
    }
    
    if (ret != 0)
    {
        if (argc == optind || argc > optind + 2)
        {
            // no or more than 2 arguments left unprocessed
            error_message("wrong number of arguments; execute " T2C_SHORT_NAME " --help for usage summary"); 
            ret = 0;
        }
        else 
        {
            // the first non-option argument should be the subsuite root directory
            psettings->suite_root_dir = mist_path_absolute(argv[optind]);
            if (psettings->suite_root_dir == NULL)
            {
                error_message("unable to obtain the absolute path to the subsuite root directory");
                cleanup_settings(psettings);
                exit(EXIT_FAILURE);
            }
            
            ++optind; // to the next argument (if any)
            if (argc == optind + 1)
            {
                // .t2c file path is specified
                psettings->t2c_file_path = mist_path_absolute(argv[optind]);
                if (psettings->t2c_file_path == NULL)
                {
                    error_message("unable to obtain the absolute path to the t2c file");
                    cleanup_settings(psettings);
                    exit(EXIT_FAILURE);
                }
            }
        }

    }
    return ret;
}

void
error_message(const char* msg)
{
    assert(msg != NULL);
    print_error(msg);
    return;
}

void 
error_message_for_mist_code(EMistErrorCode ec)
{
    switch (ec)
    {
    case MIST_OUT_OF_MEMORY:
        error_message("not enough memory");
        break;
    
    case MIST_OPEN_FILE_FAILED:
        error_message("unable to open file");
        break;
        
    case MIST_READ_FILE_FAILED:
        error_message("unable to read data from file");
        break;
        
    case MIST_WRITE_FILE_FAILED:
        error_message("unable to write data to file");
        break;
    
    default:
        break;
    }
    return;
}

int
init(int argc, char* argv[], struct SSettings* psettings)
{
    assert(psettings != NULL);
    
    // A subdirectory of the t2c data directory containing default templates.
    static const char* default_template_dir = "templates";

    if (load_args(argc, argv, psettings) == 0)
    {
        return 0;
    }
    
    // Get data directory
    psettings->t2c_data_dir = get_data_dir();
    if (psettings->t2c_data_dir == NULL)
    {
        print_error(errFailedToConstructDataPath);
        return 0;
    }
    
    // Check if the data directory exists
    if (!mist_dir_exists(psettings->t2c_data_dir))
    {
        print_error(errNoDataDir, psettings->t2c_data_dir);
        return 0;
    }
    
    if (mist_dir_is_root(psettings->t2c_data_dir))
    {
        print_error(errDataDirIsRoot);
        return 0;
    }
        
    // If template directory is not specified, use the default one.
    if (psettings->template_dir == NULL)
    {
        psettings->template_dir = mist_path_sum(psettings->t2c_data_dir, default_template_dir);
        if (psettings->template_dir == NULL)
        {
            error_message("unable to construct path to the template directory");
            return 0;
        }
    }
    
    // Check if the template directory exists
    if (!mist_dir_exists(psettings->template_dir))
    {
        print_error(errNoTemplateDir, psettings->template_dir);
        return 0;
    }
    
    if (mist_dir_is_root(psettings->template_dir))
    {
        error_message("root directory should not be used as a template directory");
        return 0;
    }
    
    // Check if the suite root directory exists
    if (!mist_dir_exists(psettings->suite_root_dir))
    {
        print_error(errNoSuiteDir, psettings->suite_root_dir);
        return 0;
    }
    
    if (mist_dir_is_root(psettings->suite_root_dir))
    {
        error_message("root directory should not be used as a test suite root directory");
        return 0;
    }
    
    // If .t2c file is specified, check whether it exists
    if (psettings->t2c_file_path != NULL && !mist_file_exists(psettings->t2c_file_path))
    {
        print_error(errNoFile, psettings->t2c_file_path);
        return 0;
    }
    
    psettings->test_group_params = smap_create();
    if (psettings->test_group_params == NULL)
    {
        error_message("unable to create string map");
        return 0;
    }
    
    return 1;
}

char*
get_data_dir()
{
    char* res = NULL;
    
    // If $T2C_DATADIR variable is defined, we'll use it.
    char* tmp = getenv("T2C_DATADIR");
    if (tmp != NULL && strlen(tmp) != 0)
    {
        res = mist_path_absolute(tmp);
        return res;
    }
    
    tmp = getenv("APPDATA");
    if (tmp != NULL && strlen(tmp) != 0)
    {
        res = mist_path_sum(tmp, T2C_SHORT_NAME);
        if (res == NULL)
        {
            return NULL;
        }
        res = mist_path_to_unix_slashes(res);
        
        if (mist_dir_exists(res))
        {
            return res;
        }
        else
        {
            free(res);
        }
    }
        
    res = mist_path_absolute(DATADIR);
    return res;
}

int
load_configuration(struct SSettings* psettings)
{
    assert(psettings != NULL);
    static char pname_target_key[] = PNAME_TARGET;
    static char pname_tpl_dir_key[] = PNAME_TEMPLATE_DIR;
    
    // Check if .cfg file is present in the suite root directory and load it.
    char* err = NULL;
    EMistErrorCode ec = mist_load_config_file_from_dir(psettings->suite_root_dir, 
        psettings->test_group_params, &err);
    if (ec != MIST_OK)
    {
        if (err != NULL)
        {
            print_error(errFailedToLoadCfgWithErr, psettings->suite_root_dir, err);
            free(err);
        }
        else
        {
            print_error(errFailedToLoadCfg, psettings->suite_root_dir);
        }
        return 0;
    }
    assert(err == NULL);
    
    const char* key = smap_check_duplicate_keys(psettings->test_group_params);
    if (key != NULL)
    {
        print_error(errDupParam, key);
        return 0;
    }
    
    if (psettings->target != NULL)
    {
        // Update the collection of parameters
        if (smap_set_value(psettings->test_group_params, 
                pname_target_key, psettings->target) == 0)
        {
            error_message_for_mist_code(MIST_OUT_OF_MEMORY);
            return 0;   
        }
    }
    else
    {
        // Get the name of the target to generate files for.
        const char* tstr = smap_lookup(psettings->test_group_params, pname_target_key);
        if (tstr == NULL)
        {
            print_error(errMissingTarget, pname_target_key);
            return 0;
        }
        
        psettings->target = (char*)strdup(tstr);
        if (psettings->target == NULL)
        {
            error_message_for_mist_code(MIST_OUT_OF_MEMORY);
            return 0;  
        }
    }
    
    if (mist_name_is_bad(psettings->target))
    {
        print_error(errInvalidTargetName, psettings->target);
        return 0;
    }
    
    char* full_target = mist_path_sum(psettings->target, 
        (psettings->t2c_file_path == NULL ? "common" : "test"));
    if (full_target == NULL)
    {
        error_message_for_mist_code(MIST_OUT_OF_MEMORY);
        return 0;
    }
    
    char* tpath = mist_path_sum(psettings->template_dir, full_target);
    if (tpath == NULL)
    {
        error_message_for_mist_code(MIST_OUT_OF_MEMORY);
        free(full_target);
        return 0;
    }
    
    free(psettings->template_dir);
    psettings->template_dir = tpath;
        
    // Check if an auxiliary template directory is specified for the suite 
    // and construct its absolute path
    const char* aux_rel_path = smap_lookup(psettings->test_group_params, pname_tpl_dir_key);
    if (aux_rel_path != NULL)
    {
        if (aux_rel_path[0] == '\0')
        {
            // TEMPLATE_DIR parameter is defined but has empty value.
            error_message(errTemplateDirEmpty);
            free(full_target);
            return 0;
        }
        
        assert(psettings->suite_root_dir != NULL);
        assert(mist_path_is_absolute(psettings->suite_root_dir));
        
        tpath = mist_path_sum(psettings->suite_root_dir, aux_rel_path);
        if (tpath == NULL)
        {
            error_message_for_mist_code(MIST_OUT_OF_MEMORY);
            free(full_target);
            return 0;
        }
        
        psettings->aux_template_dir = mist_path_absolute(tpath);
        free(tpath);
        tpath = NULL;
        
        if (psettings->aux_template_dir == NULL)
        {
            error_message_for_mist_code(MIST_OUT_OF_MEMORY);
            free(full_target);
            return 0;
        }
        
        if (!mist_dir_exists(psettings->aux_template_dir))
        {
            print_error(errNoAuxTemplateDir, psettings->aux_template_dir);
            free(full_target);
            return 0;
        }
        
        tpath = mist_path_sum(psettings->aux_template_dir, full_target);
        if (tpath == NULL)
        {
            error_message_for_mist_code(MIST_OUT_OF_MEMORY);
            free(full_target);
            return 0;
        }
        
        free(psettings->aux_template_dir);
        psettings->aux_template_dir = tpath;
    }
    
    // If a .t2c file is to be processed, get its name, check if there is 
    // a custom template dir and/or a custom configuration file for it. 
    if (psettings->t2c_file_path != NULL)
    {
        static const char t2c_ext[] = ".t2c";
        static const size_t t2c_ext_len = sizeof(t2c_ext) / sizeof(t2c_ext[0]) - 1;
        
        // get the name of the file and strip ".t2c" extension,
        // leave other extensions intact
        char* name = mist_path_get_last(psettings->t2c_file_path);
        if (name == NULL)
        {
            error_message_for_mist_code(MIST_OUT_OF_MEMORY);
            free(full_target);
            return 0;
        }
        
        size_t len = strlen(name);
        if (len > t2c_ext_len && 
            !strncmp(&name[len - t2c_ext_len], t2c_ext, t2c_ext_len))
        {
            name[len - t2c_ext_len] = '\0';
        }
        psettings->t2c_file_name = name;
        
        // check if there are custom templates for the T2C file
        char* tpath = template_dir_for_file_path(psettings->t2c_file_path);
        if (tpath == NULL)
        {
            free(full_target);
            error_message_for_mist_code(MIST_OUT_OF_MEMORY);
            return 0;
        }
        
        assert(psettings->per_test_template_dir == NULL);
        if (mist_dir_exists(tpath) && mist_subdir_exists(tpath, full_target))
        {
            psettings->per_test_template_dir = mist_path_sum(tpath, full_target);
            if (psettings->per_test_template_dir == NULL)
            {
                free(full_target);
                free(tpath);
                error_message_for_mist_code(MIST_OUT_OF_MEMORY);
                return 0;
            }
        }
        
        free(tpath);
        
        // check if there is a custom configuration file for the T2C file
        char* base_dir = mist_path_get_containing_dir(psettings->t2c_file_path);
        if (base_dir == NULL)
        {
            free(full_target);
            error_message_for_mist_code(MIST_OUT_OF_MEMORY);
            return 0;
        }
        
        // create a temporary storage for parameters
        CStringMap* sm_upd = smap_create();
        if (sm_upd == NULL)
        {
            free(base_dir);
            free(full_target);
            error_message_for_mist_code(MIST_OUT_OF_MEMORY);
            return 0;
        }
        
        ec = mist_load_config_file_for_name(base_dir, psettings->t2c_file_name,
            sm_upd, &err);
        
        if (ec == MIST_OK) 
        {
            // there is a custom .cfg file and it has been successfully loaded
            key = smap_check_duplicate_keys(sm_upd);
            if (key != NULL)
            {
                print_error(errDupParamTG, key, psettings->t2c_file_name);
                smap_destroy(sm_upd);
                free(base_dir);
                free(full_target);
                return 0;
            }
            
            // it is not allowed to specify TARGET as a parameter specific to 
            // the test group 
            if (smap_lookup(sm_upd, pname_target_key) != NULL)
            {
                print_error(errInvalidParamTG, pname_target_key, psettings->t2c_file_name);
                smap_destroy(sm_upd);
                free(base_dir);
                free(full_target);
                return 0;
            }
            
            if (smap_update(psettings->test_group_params, sm_upd) == 0)
            {
                smap_destroy(sm_upd);
                free(base_dir);
                free(full_target);
                error_message_for_mist_code(MIST_OUT_OF_MEMORY);
                return 0;
            }
        }
        else if (ec == MIST_OPEN_FILE_FAILED) 
        {
            // no custom .cfg file or bad permissions => do nothing, just go on
            free(err); // just in case
        }
        else 
        {
            // some other error occured
            if (err != NULL)
            {
                print_error(errFailedToLoadTGCfgWithErr, 
                    psettings->t2c_file_name, base_dir, err);
                free(err);
            }
            else
            {
                print_error(errFailedToLoadTGCfg, psettings->t2c_file_name, base_dir);
            }
            
            free(base_dir);
            smap_destroy(sm_upd);
            free(full_target);
            return 0;
        }
        
        free(base_dir);
        smap_destroy(sm_upd);
    }
    
    free(full_target);
    
    // add standard parameters to the string map
    if (set_std_params(psettings) == 0)
    {
        print_error(errFailedToSetStdParams);
        return 0;
    }
    
    return 1;
}

void 
cleanup_settings(struct SSettings* psettings)
{
    assert(psettings != NULL);
    
    free(psettings->template_dir);
    free(psettings->suite_root_dir);
    free(psettings->t2c_file_path);
    free(psettings->t2c_file_name);
    free(psettings->target);
    
    if (psettings->test_group_params != NULL)
    {
        smap_destroy(psettings->test_group_params);
    }
    
    free(psettings->t2c_data_dir);
    free(psettings->aux_template_dir);
    free(psettings->per_test_template_dir);
    
    if (psettings->template_list != NULL)
    {
        smap_destroy(psettings->template_list);
    }
    
    free(psettings->test_name_file_path);
    
    return;
}

char* 
template_dir_for_file_path(const char* file_path)
{
    assert(file_path != NULL);
    assert(file_path[0] != '\0');
    assert(mist_file_exists(file_path));
    
    static const char t2c_ext[] = ".t2c";
    static const size_t t2c_ext_len = sizeof(t2c_ext) / sizeof(t2c_ext[0]) - 1;
    static const char tpl_ext[] = ".templates";
    static const size_t tpl_ext_len = sizeof(tpl_ext) / sizeof(tpl_ext[0]) - 1;
    
    size_t len = strlen(file_path);
    size_t pos = len;
    
    if (len > t2c_ext_len && 
        !strncmp(&file_path[len - t2c_ext_len], t2c_ext, t2c_ext_len))
    {
        pos = len - t2c_ext_len;
    }
    
    char* res = malloc(pos + tpl_ext_len + 1);
    if (res == NULL)
    {
        return NULL;
    }
    
    strncpy(res, file_path, pos);
    strncpy(&res[pos], tpl_ext, tpl_ext_len);
    res[pos + tpl_ext_len] = '\0';
    
    return res;
}

int
update_template_list(const char* dir_path, struct CStringMap_* smap)
{
    assert(dir_path != NULL);
    assert(dir_path[0] != '\0');
    assert(smap != NULL);
    
    if (mist_dir_exists(dir_path))
    {
        DIR* dir = opendir(dir_path);
        if (dir == NULL)
        {
            print_error(errDirReadFailed, dir_path);
            return 0;
        }
        
        struct dirent* entry = readdir(dir);
        while (entry != NULL)
        {
            if (!mist_dir_is_special(entry->d_name))
            {
                char* path = mist_path_sum(dir_path, entry->d_name);
                if (path == NULL)
                {
                    closedir(dir);
                    return 0;
                }
                
                if (mist_dir_exists(path))
                {
                    if (smap_set_value(smap, entry->d_name, path) == 0)
                    {
                        closedir(dir);
                        free(path);
                        return 0;
                    }
                }
                
                free(path);
            }
            entry = readdir(dir);
        }
        
        closedir(dir);
    }
    return 1; 
}

struct CStringMap_*
load_template_list(struct SSettings* psettings)
{
    assert(psettings != NULL);
    assert(psettings->template_dir != NULL);
    
    struct CStringMap_* tpl_map = smap_create();
    if (tpl_map == NULL)
    {
        return NULL;
    }
    
    // Traverse main template directory and populate the list
    DIR* dir = opendir(psettings->template_dir);
    if (dir != NULL)
    {
        struct dirent* entry = readdir(dir);
        while (entry != NULL)
        {
            if (!mist_dir_is_special(entry->d_name))
            {
                char* path = mist_path_sum(psettings->template_dir, entry->d_name);
                if (path == NULL)
                {
                    smap_destroy(tpl_map);
                    closedir(dir);
                    return NULL;
                }
                
                if (mist_dir_exists(path))
                {
                    if (smap_add_element(tpl_map, entry->d_name, path) == 0)
                    {
                        smap_destroy(tpl_map);
                        closedir(dir);
                        free(path);
                        return NULL;
                    }
                }
                
                free(path);
            }
            entry = readdir(dir);
        }
        closedir(dir);
    }
    
    // Check other template directories and update the list of templates
    if (psettings->aux_template_dir != NULL)
    {
        if (update_template_list(psettings->aux_template_dir, tpl_map) == 0)
        {
            print_error(errReadTplListFailed, psettings->aux_template_dir);
            smap_destroy(tpl_map);
            return NULL;
        }
    }
    
    if (psettings->t2c_file_path != NULL && 
        psettings->per_test_template_dir != NULL)
    {
        if (update_template_list(psettings->per_test_template_dir, tpl_map) == 0)
        {
            print_error(errReadTplListFailed, psettings->per_test_template_dir);
            smap_destroy(tpl_map);
            return NULL;
        }
    }
    
    if (smap_get_size(tpl_map) == 0)
    {
        print_error(errNoTemplates, psettings->target);
        smap_destroy(tpl_map);
        return NULL;
    }
    
    assert(smap_check_duplicate_keys(tpl_map) == NULL);
    return tpl_map;
}

int 
set_std_params(struct SSettings* psettings)
{
    assert(psettings != NULL);
    assert(psettings->test_group_params != NULL);
    assert(psettings->suite_root_dir != NULL);
    
    static char datadir_key[]        = "T2C_DATADIR";
    static char suite_root_dir_key[] = "SUITE_ROOT_DIR";
    static char suite_name_key[]     = "SUITE_NAME";
    static char t2c_file_name_key[]  = "T2C_FILE_NAME";
    static char year_key[]           = "YEAR";
        
    static const char suffix[] = "-t2c";
    static const size_t suf_len = sizeof(suffix) / sizeof(suffix[0]) - 1;
    
    if (smap_lookup(psettings->test_group_params, &datadir_key[0]) != NULL)
    {
        print_error(errValueAlreadySpecified, &datadir_key[0]);
        return 0;
    }
    
    if (smap_lookup(psettings->test_group_params, &suite_root_dir_key[0]) != NULL)
    {
        print_error(errValueAlreadySpecified, &suite_root_dir_key[0]);
        return 0;
    }
    
    if (smap_lookup(psettings->test_group_params, &suite_name_key[0]) != NULL)
    {
        print_error(errValueAlreadySpecified, &suite_name_key[0]);
        return 0;
    }
    
    if (smap_lookup(psettings->test_group_params, &t2c_file_name_key[0]) != NULL)
    {
        print_error(errValueAlreadySpecified, &t2c_file_name_key[0]);
        return 0;
    }
    
    if (smap_lookup(psettings->test_group_params, &year_key[0]) != NULL)
    {
        print_error(errValueAlreadySpecified, &year_key[0]);
        return 0;
    }
    
    // T2C_DATADIR
    if (!smap_add_element(psettings->test_group_params, &datadir_key[0], psettings->t2c_data_dir))
    {
        return 0;
    }
    
    // SUITE_ROOT_DIR
    if (!smap_add_element(psettings->test_group_params, &suite_root_dir_key[0], psettings->suite_root_dir))
    {
        return 0;
    }
    
    // SUITE_NAME
    char* suite_name = mist_path_get_last(psettings->suite_root_dir);
    if (suite_name == NULL)
    {
        return 0;
    }
    size_t len = strlen(suite_name);
    if (len > suf_len && 
        !strncmp(&suite_name[len - suf_len], suffix, suf_len))
    {
        suite_name[len - suf_len] = '\0';
    }
    
    int bOK = smap_add_element(psettings->test_group_params, &suite_name_key[0], suite_name);
    free(suite_name);
    if (!bOK)
    {
        return 0;
    }
    
    // T2C_FILE_NAME
    if (psettings->t2c_file_path != NULL)
    {
        assert(psettings->t2c_file_name != NULL);
        if (smap_add_element(psettings->test_group_params, 
                &t2c_file_name_key[0], psettings->t2c_file_name) == 0)
        {
            return 0;
        }
    }
    
    // YEAR
    static char year[MAX_NUM_DIGITS];
    
    time_t raw_tm = time(NULL);
    struct tm* local_tm = localtime(&raw_tm);
    
    if (local_tm != NULL)
    {
        sprintf(&year[0], "%d", local_tm->tm_year + 1900);
        if (!smap_add_element(psettings->test_group_params, &year_key[0], &year[0]))
        {
            return 0;
        }
    }

    // TEST_MODULE
    if (!populate_test_module_list(psettings))
    {
        return 0;
    }
    
    return 1;
}

int
generate_files(struct SSettings* psettings)
{
    assert(psettings != NULL);

    // name of the template for the individual tests
    static const char* test_case_tpl = "test_case";
    
    // name of the template for the groups of tests
    static const char* test_group_tpl = "test_group";
    
    CStringMap* tparam = NULL;  // test-specific parameters
    
    int common_mode = (psettings->t2c_file_path == NULL);
    if (!common_mode)
    {
        // load .t2c file, process "test_case" template group, prepare a string map
        // with parameters for "test_group"
        tparam = process_t2c_file(psettings, test_case_tpl);
        if (tparam == NULL)
        {
            print_error(errGenTestsFail, psettings->t2c_file_path);
            return 0;
        }
    }

    size_t num = smap_get_size(psettings->template_list);
    TStringPair** base = smap_as_array(psettings->template_list);
    
    for (size_t i = 0; i < num; ++i) // for each template to be processed...
    {
        // for each template except "test_case" (already processed) ...
        if (common_mode || strcmp(base[i]->key, test_case_tpl) != 0)
        {
            EMistErrorCode ec = MIST_OK;
       
            char* error = NULL;
            CMistTemplateGroup* tg = NULL;
            CMistTemplateGroup* path_tg = NULL;
            
            const char* dir = base[i]->val;
            assert(dir != NULL);
            
            // load the templates
            ec = mist_tg_load_from_dir_impl(dir, &tg, &path_tg, &error);
            if (ec != MIST_OK)
            {
                if (error != NULL)
                {
                    print_error("%s", error);
                    free(error);
                }
                
                if (tparam != NULL)
                {
                    smap_destroy(tparam);
                }
                
                print_error(errTplLoadFail, dir);
                return 0;
            }
            
            // create path to the output file
            const char* path = mist_tg_generate_path_string_impl(path_tg, 
                psettings->test_group_params, &error);
            if (path == NULL)
            {
                if (error != NULL)
                {
                    print_error("%s", error);
                    free(error);
                }
                
                print_error(errOutFileConstructFail);
                mist_tg_destroy_impl(tg);
                mist_tg_destroy_impl(path_tg);
                
                if (tparam != NULL)
                {
                    smap_destroy(tparam);
                }
                return 0;
            }
            
            // If we are processing "test_group" template, set test-specific parameters
            // for it. The common ones ('psettings->test_group_params') will be set later
            // by mist_tg_generate_file_impl() - see below.
            if (!common_mode && !strcmp(base[i]->key, test_group_tpl))
            {
                // add the name of the output file as a parameter
                int bOK = smap_add_element(tparam, "this_file", path);
                               
                // set test-specific parameters for 'tg'.
                ec = mist_tg_set_values_impl(tg, tparam);
                if (!bOK || ec != MIST_OK)
                {
                    print_error(errFailedToSetTestParams, psettings->t2c_file_path);
                    
                    mist_tg_destroy_impl(tg);
                    mist_tg_destroy_impl(path_tg);
                    
                    if (tparam != NULL)
                    {
                        smap_destroy(tparam);
                    }
                    return 0;
                }
            }
            
            // generate the output file
            ec = mist_tg_generate_file_impl(tg, path, psettings->test_group_params, &error);
            if (ec != MIST_OK)
            {
                if (error != NULL)
                {
                    print_error("%s", error);
                    free(error);
                }
                print_error(errGenOutputFail, dir);
                
                mist_tg_destroy_impl(tg);
                mist_tg_destroy_impl(path_tg);
                
                if (tparam != NULL)
                {
                    smap_destroy(tparam);
                }
                return 0;
            }
            
            mist_tg_destroy_impl(tg);
            mist_tg_destroy_impl(path_tg);
        }
    }
    
    if (tparam != NULL)
    {
        smap_destroy(tparam);
    }
    return 1;
}

CStringMap*
process_t2c_file(struct SSettings* psettings, const char* tc_group_name)
{
    assert(psettings != NULL);
    assert(psettings->t2c_file_path != NULL);
    assert(tc_group_name != NULL);
    
    CStringMap* params = smap_create();
    if (params == NULL)
    {
        return NULL;
    }
    
    if (!smap_add_element(params, "t2c_file", psettings->t2c_file_path))
    {
        smap_destroy(params);
        return NULL;
    }
    
    // Load the template group for test cases
    CMistTemplateGroup* tg = load_tc_template_group(psettings, tc_group_name);
    if (tg == NULL)
    {
        smap_destroy(params);
        return NULL;
    }

    // load the .t2c file and prepare parameters for the test.
    char* contents = t2c_read_file(psettings->t2c_file_path);
    if (contents == NULL)
    {
        print_error(errT2CFileLoadFailed, psettings->t2c_file_path);
                    
        mist_tg_destroy_impl(tg);
        smap_destroy(params);
        return NULL;
    }
    
    char* error = NULL;
    int bOK = t2c_load_data(contents, params, tg, &error);
    if (!bOK)
    {
        if (error != NULL)
        {
            print_error("%s", error);
            free(error);
        }
        smap_destroy(params);
    }
    
    free(contents);
    mist_tg_destroy_impl(tg);
    return ((bOK) ? params : NULL);
}

CMistTemplateGroup*
load_tc_template_group(struct SSettings* psettings, const char* tc_group_name)
{
    assert(psettings != NULL);
    assert(psettings->t2c_file_path != NULL);
    assert(tc_group_name != NULL);
    
    // 1. Lookup the path to the template group directory
    char* tg_name = (char*)strdup(tc_group_name);
    if (tg_name == NULL)
    {
        return NULL;
    }
    
    const char* tg_dir = smap_lookup(psettings->template_list, tg_name);
    free(tg_name);
    
    if (tg_dir == NULL)
    {
        print_error(errTGNotExist, tc_group_name);
        return NULL;
    }
    
    // 2. Try to load the template group
    EMistErrorCode ec = MIST_OK;
    char* error = NULL;
    CMistTemplateGroup* tg = NULL;
    CMistTemplateGroup* path_tg = NULL;
    
    ec = mist_tg_load_from_dir_impl(tg_dir, &tg, &path_tg, &error);
    if (ec != MIST_OK)
    {
        if (error != NULL)
        {
            print_error("%s", error);
            free(error);
        }
        print_error(errTplLoadFail, tg_dir);
        return NULL;
    }
    
    mist_tg_destroy_impl(path_tg); // we don't need this group
    
    return tg;
}

int
t2c_load_data(const char* source, CStringMap* params, 
               CMistTemplateGroup* tc_group, char** error_descr)
{
    assert(source != NULL);
    assert(params != NULL);
    assert(tc_group != NULL);
    
    if (error_descr != NULL)
    {
        *error_descr = NULL;
    }
    
    CT2CParser parser;
    t2c_parser_init(&parser, source, source, source + strlen(source));
    
    EParserResult pr;

    // parse the t2c source:
    // 1. header (optional)
    pr = t2c_match_header(&parser);
    if (pr != PR_MATCHED && pr != PR_NOT_MATCHED)
    {
        t2c_set_error(error_descr, &parser);
        return 0;
    }
    
    if (pr == PR_MATCHED)
    {
        if (t2c_parse_header(&parser, params) != PR_PARSE_OK)
        {
            t2c_set_error(error_descr, &parser);
            return 0;
        }
        
        t2c_parser_advance(&parser);
    }
    
    // This string map is to be passed to t2c_match_section() for the sections
    // that must have no attributes - just to check if they actually do have
    // them. Error will be reported if this is the case.
    CStringMap* tmp_sm = NULL;
    
    // 2. <GLOBAL> section (optional)
    pr = t2c_match_section(&parser, sec_global, &tmp_sm);
    if (pr != PR_MATCHED && pr != PR_NOT_MATCHED)
    {
        t2c_set_error(error_descr, &parser);
        return 0;
    }
    if (tmp_sm != NULL)
    {
        smap_destroy(tmp_sm);
        t2c_format_parse_error(&parser, errUnexpectedAttribs, sec_global);
        t2c_set_error(error_descr, &parser);
        return 0;
    }
    
    if (pr == PR_MATCHED)
    {
        if (t2c_parse_simple_section(&parser, "global", params) != PR_PARSE_OK)
        {
            t2c_set_error(error_descr, &parser);
            return 0;
        }
        
        t2c_parser_advance(&parser);
    }
    
    // 3. <STARTUP> section (optional)
    assert(tmp_sm == NULL);
    pr = t2c_match_section(&parser, sec_startup, &tmp_sm);
    if (pr != PR_MATCHED && pr != PR_NOT_MATCHED)
    {
        t2c_set_error(error_descr, &parser);
        return 0;
    }
    if (tmp_sm != NULL)
    {
        smap_destroy(tmp_sm);
        t2c_format_parse_error(&parser, errUnexpectedAttribs, sec_startup);
        t2c_set_error(error_descr, &parser);
        return 0;
    }
    
    if (pr == PR_MATCHED)
    {
        if (t2c_parse_simple_section(&parser, "startup", params) != PR_PARSE_OK)
        {
            t2c_set_error(error_descr, &parser);
            return 0;
        }
        
        t2c_parser_advance(&parser);
    }
    
    // 4. <CLEANUP> section (optional)
    assert(tmp_sm == NULL);
    pr = t2c_match_section(&parser, sec_cleanup, &tmp_sm);
    if (pr != PR_MATCHED && pr != PR_NOT_MATCHED)
    {
        t2c_set_error(error_descr, &parser);
        return 0;
    }
    if (tmp_sm != NULL)
    {
        smap_destroy(tmp_sm);
        t2c_format_parse_error(&parser, errUnexpectedAttribs, sec_cleanup);
        t2c_set_error(error_descr, &parser);
        return 0;
    }
    
    if (pr == PR_MATCHED)
    {
        if (t2c_parse_simple_section(&parser, "cleanup", params) != PR_PARSE_OK)
        {
            t2c_set_error(error_descr, &parser);
            return 0;
        }
        
        t2c_parser_advance(&parser);
    }
    
    // 5. <BLOCK> sections (at least one should be present)
    
    // index of a block is 1-based, 0 means that no blocks have been found so far
    int block_index = 0;
    
    // absolute index of a test generated from the .t2c file (1-based)
    int abs_test_index = 0; 
    
    do
    {
        CStringMap* attr = NULL;
        
        pr = t2c_match_section(&parser, sec_block, &attr);
        if (pr != PR_MATCHED)
        {
            if (pr == PR_NOT_MATCHED)
            {
                parser.beg = mist_find_in_range_first_not_of(parser.beg, parser.end, wspace, sz_wspace);
                t2c_format_parse_error(&parser, errBlockExpected, NULL);
            }
            t2c_set_error(error_descr, &parser);
            assert(attr == NULL);
            
            return 0;
        }
        
        ++block_index;
        
        // parse the block
        pr = t2c_parse_block_section(&parser, block_index, &abs_test_index, params, tc_group, attr);
        if (pr != PR_PARSE_OK)
        {
            t2c_set_error(error_descr, &parser);
            if (attr != NULL)
            {
                smap_destroy(attr);
            }
            
            return 0;
        }
        
        t2c_parser_advance(&parser);
        parser.beg = mist_find_in_range_first_not_of(parser.beg, parser.end, wspace, sz_wspace);
        
        if (attr != NULL)
        {
            smap_destroy(attr);
        }
    }
    while(parser.beg < parser.end);
    
    // set "num_tests" parameter as the last value of 'abs_test_index'
    assert(abs_test_index > 0); // if we got here, there should have been at least one test
    
    char snum[MAX_NUM_DIGITS + 1];
    sprintf(snum, "%d", abs_test_index);
    
    int bOK = smap_add_element(params, "num_tests", &snum[0]);
    if (!bOK)
    {
        return 0;
    }
    
    return 1;
}

char*
t2c_read_file(const char* path)
{
    assert(path != NULL);
    
    FILE* fd = fopen(path, "r");
    if (fd == NULL)
    {
        return NULL;
    }
    
    int sz = (int)(mist_file_get_size(fd) + 1);
    if (sz == 0)
    {
        fclose(fd);
        return NULL;
    }
    assert(sz > 0);
    
    if (sz == 1) // the file is empty: its size is 0 bytes
    {
        print_error(errEmptyFile, path);
        fclose(fd);
        return NULL;
    }
    
    char* contents = (char*)malloc(sz);
    if (contents == NULL)
    {
        fclose(fd);
        return NULL;
    }
    
    // current position in the buffer
    char* pos = contents;
    
    int bOK = 1;
    do
    {
        char* res = fgets(pos, sz, fd);
        
        if (ferror(fd))
        {
            bOK = 0;
            break;
        }
        
        if (res != NULL)
        {
            size_t len = strlen(pos);
            if (!t2c_is_comment(pos, pos + len))
            {
                pos += len;
            }
            else
            {
                *pos = '\n'; // necessary to count lines properly
                ++pos;
            }
        }
        
        *pos = '\0';
    }
    while (!feof(fd));
    
    fclose(fd);
    if (bOK)
    {
        return contents;
    }
    else
    {
        free(contents);
        return NULL;
    }
}

int
t2c_is_directive(const char* beg, const char* end)
{
    static const char* dir_names[] = {
    // C preprocessor directives (according to C99 standard - 
    // ISO/IEC 9899:1999 (E), section 6.10):
        "if",
        "ifdef",
        "ifndef",
        "elif",
        "else",
        "endif",
        "include",
        "define",
        "undef",
        "line",
        "error",
        "pragma",
    
    // A few C# preprocessor directives (see "C# Language Reference" in MSDN),  
    // namely, all those that are not used by C preprocessor and, hence, are not 
    // listed above:
        "warning",
        "region",
        "endregion",
        
    // T2C directives:
        T2C_DIR_LIBRARY,
        T2C_DIR_LIBSECTION,
        T2C_DIR_ADD_RCAT
    };
    
    assert(beg != NULL);
    assert(end != NULL);
    assert(beg <= end);
    
    size_t len = end - beg;
    if (len < 2)
    {
        return 0;
    }
    
    for (size_t i = 0; i < (sizeof(dir_names) / sizeof(dir_names[0])); ++i)
    {
        size_t dlen = strlen(dir_names[i]);
        if (len >= dlen && !strncmp(beg, dir_names[i], dlen))
        {
            if (beg[dlen] == '\0' || 
                beg[dlen] == ' ' || beg[dlen] == '\t' || 
                beg[dlen] == '\n' || beg[dlen] == '\r')
            {
                return 1;
            }
        }
    }
    
    return 0;
}

int
t2c_is_comment(const char* beg, const char* end)
{
    assert(beg != NULL);
    assert(end != NULL);
    assert(beg <= end);
    
    // skip leading spaces and tabs
    const char* pos = beg;
    for (; pos < end; ++pos)
    {
        if (*pos != ' ' && *pos != '\t')
        {
            break;
        }
    }
    
    if (pos == end || *pos != '#')
    {
        return 0;
    }
    
    // skip spaces and tabs after '#'
    for (++pos; pos < end; ++pos)
    {
        if (*pos != ' ' && *pos != '\t')
        {
            break;
        }
    }
    
    if (!t2c_is_directive(pos, end))
    {
        return 1;
    }
    
    return 0;
}

EParserResult
t2c_match_header(CT2CParser* parser)
{
    assert(t2c_is_parser_valid(parser));
    
    const char* pos = mist_find_in_range_first_of(parser->beg, parser->end, "<", 1);
    if (pos == parser->end)
    {
        t2c_format_parse_error(parser, errNoSectionsFound, NULL);
        return PR_SYNTAX_ERROR;
    }
    
    if (pos[1] == '/') // found "</"
    {
        parser->beg = pos;
        t2c_format_parse_error(parser, errUnexpectedClosingTag, NULL);
        return PR_SYNTAX_ERROR;
    }
    
    if (pos != parser->beg)
    {
        parser->end = pos;
        return PR_MATCHED;
    }
    else
    {
        return PR_NOT_MATCHED;
    }
}

EParserResult
t2c_parse_header(CT2CParser* parser, CStringMap* params)
{
    assert(t2c_is_parser_valid(parser));
    assert(params != NULL);
    
    assert(parser->beg < parser->end);

    const char* pos = parser->beg;
    while (1)
    {
        pos = mist_find_in_range_first_not_of(pos, parser->end, wspace, sz_wspace);
        if (pos == parser->end)
        {
            break;
        }
        
        if (*pos != '#')
        {
            parser->beg = pos;
            t2c_format_parse_error(parser, errBadLineInHeader, NULL);
            return PR_SYNTAX_ERROR;
        }
        
        pos = mist_find_in_range_first_not_of(pos + 1, parser->end, ws, sz_ws);
        assert(pos != parser->end);
        
        const char* dend = mist_find_in_range_first_of(pos, parser->end, ws, sz_ws);
        parser->beg = dend;
        
        if (!strncmp(pos, T2C_DIR_LIBRARY, dend - pos))
        {
            // process #library directive
            EParserResult pr = t2c_parse_library_directive(parser, params);
            if (pr != PR_PARSE_OK)
            {
                return PR_SYNTAX_ERROR;
            }
        }
        else if (!strncmp(pos, T2C_DIR_LIBSECTION, dend - pos))
        {
            // process #libsection directive
            EParserResult pr = t2c_parse_libsection_directive(parser, params);
            if (pr != PR_PARSE_OK)
            {
                return PR_SYNTAX_ERROR;
            }
        }
        else if (!strncmp(pos, T2C_DIR_ADD_RCAT, dend - pos))
        {
            // process #additional_req_catalogues directive
            EParserResult pr = t2c_parse_add_rcat_directive(parser, params);
            if (pr != PR_PARSE_OK)
            {
                return PR_SYNTAX_ERROR;
            }
        }
        else
        {
            // Unknown directive
            parser->beg = pos;
            t2c_format_parse_error(parser, errBadLineInHeader, NULL);
            return PR_SYNTAX_ERROR;
        }
        
        pos = parser->beg; //'parser->beg' has been changed by t2c_parse_*_directive()
    }
    
    return PR_PARSE_OK;
}

EParserResult
t2c_parse_block_section(CT2CParser* parser, 
    int block_index, int* abs_test_index,
    CStringMap* params, 
    CMistTemplateGroup* tc_group, CStringMap* attr)
{
    assert(t2c_is_parser_valid(parser));
    assert(params != NULL);
    assert(tc_group != NULL);
    assert(block_index > 0);
    assert(abs_test_index != NULL);

    // create a parser for the contents of the block...
    CT2CParser block_parser;
    t2c_parser_init(&block_parser, parser->source, parser->beg, parser->end);
    
    // ... and a string map to hold the common parameters of the tests in the
    // block.
    CStringMap* block_params = smap_create();
    if (block_params == NULL)
    {
        return PR_NO_MEMORY;
    }
    
    EParserResult pr;
    
    // This string map is to be passed to t2c_match_section() for the sections
    // that must have no attributes - just to check if they actually do have
    // them. Error will be reported if this is the case.
    CStringMap* tmp_sm = NULL;
    
    // 1. Process <TARGETS> section (mandatory)
    pr = t2c_match_section(&block_parser, sec_targets, &tmp_sm);
    if (tmp_sm != NULL)
    {
        smap_destroy(tmp_sm);
        t2c_format_parse_error(&block_parser, errUnexpectedAttribs, sec_targets);
        t2c_set_error(&(parser->error), &block_parser);
        smap_destroy(block_params);
        return PR_SYNTAX_ERROR;
    }
    
    if (pr == PR_MATCHED)
    {
        pr = t2c_parse_targets_section(&block_parser, block_params);
        if (pr != PR_PARSE_OK)
        {
            t2c_set_error(&(parser->error), &block_parser);
            smap_destroy(block_params);
            return pr;
        }
        
        t2c_parser_advance(&block_parser);
    }
    else if (pr == PR_NOT_MATCHED)
    {
        block_parser.beg = mist_find_in_range_first_not_of(block_parser.beg,
            block_parser.end, wspace, sz_wspace);
        t2c_format_parse_error(&block_parser, errTargetsExpected, NULL);
        t2c_set_error(&(parser->error), &block_parser);

        smap_destroy(block_params);
        return PR_SYNTAX_ERROR;
    }
    else
    {
        // some other error detected
        t2c_set_error(&(parser->error), &block_parser);
        smap_destroy(block_params);
        return pr;
    }
    
    // 2. Process <DEFINE> section (optional)
    assert(tmp_sm == NULL);
    pr = t2c_match_section(&block_parser, sec_define, &tmp_sm);
    if (tmp_sm != NULL)
    {
        smap_destroy(tmp_sm);
        t2c_format_parse_error(&block_parser, errUnexpectedAttribs, sec_define);
        t2c_set_error(&(parser->error), &block_parser);
        smap_destroy(block_params);
        return PR_SYNTAX_ERROR;
    }
    if (pr != PR_MATCHED && pr != PR_NOT_MATCHED)
    {
        t2c_set_error(&(parser->error), &block_parser);
        smap_destroy(block_params);
        return pr;
    }
    
    if (pr == PR_MATCHED)
    {
        pr = t2c_parse_define_section(&block_parser, block_params);
        if (pr != PR_PARSE_OK)
        {
            t2c_set_error(&(parser->error), &block_parser);
            smap_destroy(block_params);
            return pr;
        }
        
        t2c_parser_advance(&block_parser);
    }
    
    // 3. Process <CODE> section (mandatory)
    assert(tmp_sm == NULL);
    pr = t2c_match_section(&block_parser, sec_code, &tmp_sm);
    if (tmp_sm != NULL)
    {
        smap_destroy(tmp_sm);
        t2c_format_parse_error(&block_parser, errUnexpectedAttribs, sec_code);
        t2c_set_error(&(parser->error), &block_parser);
        smap_destroy(block_params);
        return PR_SYNTAX_ERROR;
    }
    
    if (pr == PR_MATCHED)
    {
        pr = t2c_parse_simple_section(&block_parser, "code", block_params);
        if (pr != PR_PARSE_OK)
        {
            t2c_set_error(&(parser->error), &block_parser);
            smap_destroy(block_params);
            return pr;
        }
        
        t2c_parser_advance(&block_parser);
    }
    else if (pr == PR_NOT_MATCHED)
    {
        block_parser.beg = mist_find_in_range_first_not_of(block_parser.beg,
            block_parser.end, wspace, sz_wspace);
        t2c_format_parse_error(&block_parser, errCodeExpected, NULL);
        t2c_set_error(&(parser->error), &block_parser);
        
        smap_destroy(block_params);
        return PR_SYNTAX_ERROR;
    }
    else
    {
        // some other error detected
        t2c_set_error(&(parser->error), &block_parser);
        smap_destroy(block_params);
        return pr;
    }
    
    // 4. Process <FINALLY> section (optional)
    assert(tmp_sm == NULL);
    pr = t2c_match_section(&block_parser, sec_finally, &tmp_sm);
    if (tmp_sm != NULL)
    {
        smap_destroy(tmp_sm);
        t2c_format_parse_error(&block_parser, errUnexpectedAttribs, sec_finally);
        t2c_set_error(&(parser->error), &block_parser);
        smap_destroy(block_params);
        return PR_SYNTAX_ERROR;
    }
    
    if (pr != PR_MATCHED && pr != PR_NOT_MATCHED)
    {
        t2c_set_error(&(parser->error), &block_parser);
        smap_destroy(block_params);
        return pr;
    }
    
    if (pr == PR_MATCHED)
    {
        pr = t2c_parse_simple_section(&block_parser, "finally", block_params);
        if (pr != PR_PARSE_OK)
        {
            t2c_set_error(&(parser->error), &block_parser);
            smap_destroy(block_params);
            return pr;
        }
        
        t2c_parser_advance(&block_parser);
    }
    
    // 5. Prepare test case template with all parameters of the block
    // substituted and create a new template group from it.
    CMistTemplateGroup* tg = t2c_prepare_test_tg(tc_group, block_index, block_params, &(parser->error));
    if (tg == NULL)
    {
        smap_destroy(block_params);
        return (parser->error == NULL ? PR_NO_MEMORY : PR_SYNTAX_ERROR);
    }
    
    // 6. Process <VALUES> (<PURPOSE>) sections (zero or more). 
    // Use the newly created template group to generate the values of "test_case"
    // parameter (i.e. the source of each test case)
    pr = t2c_process_values_sections(&block_parser, block_index, abs_test_index, tg, params, attr);
    if (pr != PR_PARSE_OK)
    {
        t2c_set_error(&(parser->error), &block_parser);
        mist_tg_destroy_impl(tg);
        smap_destroy(block_params);
        return pr;
    }
    
    // 7. Check if unexpected text remains in the block.
    block_parser.beg = mist_find_in_range_first_not_of(block_parser.beg, block_parser.end, 
        wspace, sz_wspace);
    if (block_parser.beg != block_parser.end)
    {
        t2c_format_parse_error(&block_parser, errUnexpectedTextInBlock, NULL);
        t2c_set_error(&(parser->error), &block_parser);
        
        mist_tg_destroy_impl(tg);
        smap_destroy(block_params);
        return PR_SYNTAX_ERROR;
    }
    
    mist_tg_destroy_impl(tg);
    smap_destroy(block_params);
    
    t2c_parser_to_section_end(parser);
    return PR_PARSE_OK;
}

EParserResult
t2c_parse_library_directive(CT2CParser* parser, CStringMap* params)
{
    assert(t2c_is_parser_valid(parser));
    assert(params != NULL);
    
    static char par_name[] = "library";
    
    // check if the corresponding parameter already exists.
    if (smap_lookup(params, &par_name[0]) != NULL)
    {
        t2c_format_parse_error(parser, errParamAlreadySet, &par_name[0]);
        return PR_SYNTAX_ERROR;
    }
    
    const char* beg = mist_find_in_range_first_not_of(parser->beg, parser->end, ws, sz_ws);
    if (beg == parser->end || *beg == '\n' || *beg == '\r')
    {
        t2c_format_parse_error(parser, errDirInvalid, T2C_DIR_LIBRARY);
        return PR_SYNTAX_ERROR;
    }
    
    const char* end = mist_find_in_range_first_of(beg, parser->end, eol, sz_eol);
    char* val = mist_get_substring(beg, end);
    if (val == NULL)
    {
        return PR_NO_MEMORY;
    }
    
    int bOK = smap_add_element(params, &par_name[0], val);
    free(val);
    
    if (!bOK)
    {
        return PR_NO_MEMORY;
    }
    else
    {
        parser->beg = end;
        return PR_PARSE_OK;
    }
}

EParserResult
t2c_parse_libsection_directive(CT2CParser* parser, CStringMap* params)
{
    assert(t2c_is_parser_valid(parser));
    assert(params != NULL);
    
    static char par_name[] = "libsection";
    
    // check if the corresponding parameter already exists.
    if (smap_lookup(params, &par_name[0]) != NULL)
    {
        t2c_format_parse_error(parser, errParamAlreadySet, &par_name[0]);
        return PR_SYNTAX_ERROR;
    }
    
    const char* beg = mist_find_in_range_first_not_of(parser->beg, parser->end, ws, sz_ws);
    if (beg == parser->end || *beg == '\n' || *beg == '\r')
    {
        t2c_format_parse_error(parser, errDirInvalid, T2C_DIR_LIBSECTION);
        return PR_SYNTAX_ERROR;
    }
    
    const char* end = mist_find_in_range_first_of(beg, parser->end, eol, sz_eol);
    char* val = mist_get_substring(beg, end);
    if (val == NULL)
    {
        return PR_NO_MEMORY;
    }
    
    int bOK = smap_add_element(params, &par_name[0], val);
    free(val);
    
    if (!bOK)
    {
        return PR_NO_MEMORY;
    }
    else
    {
        parser->beg = end;
        return PR_PARSE_OK;
    }
}

EParserResult
t2c_parse_add_rcat_directive(CT2CParser* parser, CStringMap* params)
{
    assert(t2c_is_parser_valid(parser));
    assert(params != NULL);
    
    static char par_name[] = "req_cat";
    static const char delims[] = " \t,;";
    static size_t sz_delims = sizeof(delims) / sizeof(delims[0]) - 1;
    
    // more than one such directive is also allowed
    const char* beg = mist_find_in_range_first_not_of(parser->beg, parser->end, ws, sz_ws);
    if (beg == parser->end || *beg == '\n' || *beg == '\r')
    {
        t2c_format_parse_error(parser, errDirInvalid, T2C_DIR_ADD_RCAT);
        return PR_SYNTAX_ERROR;
    }
    
    const char* end = mist_find_in_range_first_of(beg, parser->end, eol, sz_eol);

    int bOK = 1;
    while (beg < end)
    {
        const char* item_end = mist_find_in_range_first_of(beg, end, delims, sz_delims);
        char* val = mist_get_substring(beg, item_end);
        if (val == NULL)
        {
            return PR_NO_MEMORY;
        }
        
        bOK = smap_add_element(params, &par_name[0], val);
        free(val);
        
        if (!bOK)
        {
            break;
        }
        
        beg = mist_find_in_range_first_not_of(item_end, end, delims, sz_delims);
    }
    
    if (!bOK)
    {
        return PR_NO_MEMORY;
    }
    else
    {
        parser->beg = end;
        return PR_PARSE_OK;
    }
}

EParserResult
t2c_parse_targets_section(CT2CParser* block_parser, CStringMap* block_params)
{
    assert(t2c_is_parser_valid(block_parser));
    assert(block_params != NULL);
    
    static char par_name[] = "target";
       
    int targets_found = 0;
    block_parser->beg = mist_find_in_range_first_not_of(block_parser->beg, block_parser->end, 
        wspace, sz_wspace);
    
    while (block_parser->beg < block_parser->end)
    {
        const char* end = mist_find_in_range_first_of(block_parser->beg, block_parser->end, 
            eol, sz_eol);
        assert(block_parser->beg < end);
                
        end = mist_find_in_range_last_not_of(block_parser->beg, end, ws, sz_ws);
        assert(end + 1 <= block_parser->end);
        ++end;
        assert(block_parser->beg < end);
        
        char* target = mist_get_substring(block_parser->beg, end);
        if (target == NULL)
        {
            return PR_NO_MEMORY;
        }
        
        int bOK = smap_add_element(block_params, &par_name[0], target);
        free(target);
        
        if (!bOK)
        {
            return PR_NO_MEMORY;
        }
        
        targets_found = 1;
        
        block_parser->beg = mist_find_in_range_first_not_of(end, block_parser->end, 
            wspace, sz_wspace);
    }
    
    if (!targets_found)
    {
        t2c_format_parse_error(block_parser, errNoTargets, NULL);
        return PR_SYNTAX_ERROR;
    }
    
    t2c_parser_to_section_end(block_parser);
    return PR_PARSE_OK;
}

EParserResult
t2c_parse_define_section(CT2CParser* block_parser, CStringMap* block_params)
{
    assert(t2c_is_parser_valid(block_parser));
    assert(block_params != NULL);
    
    static char par_name[] = "define_name";
    static char par_val[]  = "define_value";
    static const char* dname = "define";
    
    block_parser->beg = mist_find_in_range_first_not_of(block_parser->beg, block_parser->end, 
        wspace, sz_wspace);
    
    while (block_parser->beg < block_parser->end)
    {
        const char* end_line = mist_find_in_range_first_of(block_parser->beg, block_parser->end, 
            eol, sz_eol);
        assert(block_parser->beg < end_line);
        
        CMistString* def = mist_string_create_from_range(block_parser->beg, end_line);
        if (def == NULL)
        {
            return PR_NO_MEMORY;
        }
        
        mist_string_trim(def);
        
        // a record in <DEFINE> section must begin with #define
        if (def->str[0] != '#')
        {
            t2c_format_parse_error(block_parser, errBadDefine, NULL);
            mist_string_destroy(def);
            return PR_SYNTAX_ERROR;
        }
        
        const char* pos = mist_find_first_not_of(def->str + 1, ws, sz_ws);
        if (pos == NULL)
        {
            t2c_format_parse_error(block_parser, errBadDefine, NULL);
            mist_string_destroy(def);
            return PR_SYNTAX_ERROR;
        }
        
        def->str += (pos - def->str);
        pos = mist_find_first_of(def->str, ws, sz_ws);
        if (pos == NULL)
        {
            t2c_format_parse_error(block_parser, errBadDefine, NULL);
            mist_string_destroy(def);
            return PR_SYNTAX_ERROR;
        }
        
        size_t sz = pos - def->str;
        if (sz == 0 || strncmp(def->str, dname, sz) != 0)
        {
            t2c_format_parse_error(block_parser, errBadDefine, NULL);
            mist_string_destroy(def);
            return PR_SYNTAX_ERROR;
        }
        
        // after #define, there should be a name optionally followed by a value
        def->str += sz;
        mist_string_trim(def);
        
        size_t len = strlen(def->str);
        if (len == 0)
        {
            t2c_format_parse_error(block_parser, errBadDefine, NULL);
            mist_string_destroy(def);
            return PR_SYNTAX_ERROR;
        }
        
        const char* end = def->str + len;
        assert(*end == '\0');
        
        pos = mist_find_in_range_first_of(def->str, end, ws, sz_ws);
        if (pos != end)
        {
            def->str[pos - def->str] = '\0'; // isolate the name
            pos = mist_find_in_range_first_not_of(pos + 1, end, ws, sz_ws);
        }
        
        // now 'pos' points to the beginning of the value (may point to the terminating
        // null character if there is no value)
        if (!smap_add_element(block_params, &par_name[0], def->str) ||
            !smap_add_element(block_params, &par_val[0], pos))
        {
            mist_string_destroy(def);
            return PR_NO_MEMORY;
        }
        
        mist_string_destroy(def);
        
        block_parser->beg = mist_find_in_range_first_not_of(end_line, block_parser->end, 
            wspace, sz_wspace);
    }
    
    t2c_parser_to_section_end(block_parser);
    return PR_PARSE_OK;
}

EParserResult
t2c_parse_values_section(CT2CParser* block_parser, 
    int block_index, int* test_index, int* abs_test_index,
    int* num_values,
    CMistTemplateGroup* tg,
    CStringMap* params, CStringMap* attr)
{
    assert(t2c_is_parser_valid(block_parser));
    assert(block_index > 0);
    assert(test_index != NULL);
    assert(abs_test_index != NULL);
    assert(tg != NULL);
    assert(params != NULL);
    assert(num_values != NULL);
    
    CMistString* str = mist_string_create("");
    if (str == NULL)
    {
        return PR_NO_MEMORY;
    }
    
    CGrowingArray records;
    if (grar_create(&records) == 0)
    {
        mist_string_destroy(str);
        return PR_NO_MEMORY;
    }
    
    *num_values = 0;
    
    // extract the records and process them
    block_parser->beg = mist_find_in_range_first_not_of(block_parser->beg, 
        block_parser->end, wspace, sz_wspace);
    while (block_parser->beg < block_parser->end)
    {
        const char* ln_beg = block_parser->beg;
        const char* ln_end = ln_beg + 1;           

        mist_string_reset(str);
        const char* pos = NULL;
          
        int line_continues = 0;
        do
        {
            ln_end = mist_find_in_range_first_of(ln_beg, block_parser->end, eol, sz_eol);
            pos = mist_find_in_range_last_not_of(ln_beg, ln_end, wspace, sz_wspace);
            assert(pos < ln_end);
            
            // The last backslash character needs to be cut off anyway
            const char* boundary = (*pos == '\\') ? pos : (pos + 1);
            
            line_continues = (*pos == '\\');
            if ((pos > ln_beg) && (*(pos - 1) == '\\'))
            {
                // found escaped backslash
                // [NB] Only the last backslash character need to be escaped.
                line_continues = 0;
            }
            
            if (mist_string_append_range(str, ln_beg, boundary) != MIST_OK)
            {
                grar_destroy_with_elements(&records, t2c_vr_value_record_dtor, NULL);
                mist_string_destroy(str);
                return PR_NO_MEMORY;
            }
            
            ln_beg = mist_find_in_range_first_not_of(ln_end, block_parser->end, wspace, sz_wspace);
        }
        while (line_continues && ln_beg < block_parser->end);
        
        mist_string_trim(str);
        if (str->str[0] != '\0')
        {
            ++(*num_values);
                       
            // process the line
            CValueRecord* rec = NULL;
            EParserResult pr = t2c_vr_parse_value_record(str, &rec, &(block_parser->error));
            if (pr != PR_PARSE_OK)
            {
                if (block_parser->error != NULL)
                {
                    char* err = block_parser->error;
                    block_parser->error = NULL;
                    t2c_format_parse_error(block_parser, err, NULL);
                    free(err);
                }
                
                grar_destroy_with_elements(&records, t2c_vr_value_record_dtor, NULL);
                mist_string_destroy(str);
                return pr;
            }
            
            if (grar_add_element(&records, rec) == 0)
            {
                grar_destroy_with_elements(&records, t2c_vr_value_record_dtor, NULL);
                mist_string_destroy(str);
                return PR_NO_MEMORY;
            }
        }
        block_parser->beg = ln_beg;
    }
    assert((size_t)(*num_values) == grar_get_size(&records));
    
    // process the resulting 'records' array
    int bOK = t2c_parser_gen_test_sources(&records, 
        block_index, test_index, abs_test_index,
        tg, params, attr);
    
    grar_destroy_with_elements(&records, t2c_vr_value_record_dtor, NULL);
    mist_string_destroy(str);
    
    if (!bOK)
    {
        return PR_NO_MEMORY;
    }
    
    t2c_parser_to_section_end(block_parser);
    return PR_PARSE_OK;
}

int
t2c_set_index_values(CStringMap* params, 
    int block_index, int test_index, int abs_test_index)
{
    assert(params != NULL);
    char snum[MAX_NUM_DIGITS + 1];
    
    sprintf(snum, "%d", block_index);
    if (smap_add_element(params, "block_index", &snum[0]) == 0)
    {
        return 0;
    }
    
    sprintf(snum, "%d", test_index);
    if (smap_add_element(params, "test_index", &snum[0]) == 0)
    {
        return 0;
    }
    
    sprintf(snum, "%d", abs_test_index);
    if (smap_add_element(params, "abs_test_index", &snum[0]) == 0)
    {
        return 0;
    }

    return 1;
}

CMistTemplateGroup* 
t2c_prepare_test_tg(CMistTemplateGroup* tc_group, int block_index, CStringMap* block_params,
    char** error_descr)
{
    assert(tc_group != NULL);
    assert(block_params != NULL);
    assert(block_index > 0);
    
    if (error_descr != NULL)
    {
        *error_descr = NULL;
    }
    
    // Reset the template group to be used to create sources of the tests.
    mist_tg_clear_values_impl(tc_group);
    
    EMistErrorCode ec = mist_tg_set_values_impl(tc_group, block_params);
    if (ec != MIST_OK)
    {
        return NULL;
    }
    
    CGrowingArray* vals = mist_tg_evaluate_impl(tc_group);
    if (vals == NULL)
    {
        return NULL;
    }
    assert(grar_get_size(vals) > 0);
    
    if (grar_get_size(vals) > 1)
    {
        if (error_descr != NULL)
        {
            *error_descr = (char*)strdup(errTCMultivalued);
            // no matter if strdup() returns NULL now
        }
        return NULL;
    }
    
    char* source = grar_get_element(vals, char*, 0);
    assert(source != NULL);
    
    static char tpl_name[] = TC_TPL_NAME;
    
    char* err = NULL;
    CMistTemplateGroup* tg = mist_tg_create_single_impl(&tpl_name[0], source, "<%", "%>", &err);
    if (tg == NULL)
    {
        if (error_descr != NULL)
        {   
            if (err != NULL)
            {
                size_t len = strlen(err);
                *error_descr = (char*)malloc(strlen(errTplGenFail) + MAX_NUM_DIGITS + len + 1);
                if (*error_descr != NULL)
                {
                    sprintf(*error_descr, errTplGenFail, block_index, err);
                }
            }
            else
            {
                *error_descr = (char*)malloc(strlen(errTplGenFailNoMem) + MAX_NUM_DIGITS + 1);
                if (*error_descr != NULL)
                {
                    sprintf(*error_descr, errTplGenFailNoMem, block_index);
                }
            }
        }
        free(err);
        return NULL;
    }
    
    return tg; 
}

EParserResult
t2c_process_values_sections(CT2CParser* block_parser, int block_index, int* abs_test_index,
    CMistTemplateGroup* tg, CStringMap* params, CStringMap* attr)
{
    assert(t2c_is_parser_valid(block_parser));
    assert(block_index > 0);
    assert(abs_test_index != NULL);
    assert(tg != NULL);
    assert(params != NULL);
    
    int test_index = 0; // index of the test case in the block (1-based, 0 means error)
    int num_values = 0;
    
    int counter = 0;
    
    // This string map is to be passed to t2c_match_section() for the sections
    // that must have no attributes - just to check if they actually do have
    // them. Error will be reported if this is the case.
    CStringMap* tmp_sm = NULL;
    
    do
    {
        EParserResult pr;
        ++counter;
        
        // try to match <VALUES> first
        assert(tmp_sm == NULL);
        pr = t2c_match_section(block_parser, sec_values, &tmp_sm);
        if (tmp_sm != NULL)
        {
            smap_destroy(tmp_sm);
            t2c_format_parse_error(block_parser, errUnexpectedAttribs, sec_values);
            return PR_SYNTAX_ERROR;
        }
        
        if (pr == PR_NOT_MATCHED)
        {
            // try to match <PURPOSE> (T2C v.1.x style)
            assert(tmp_sm == NULL);
            pr = t2c_match_section(block_parser, sec_purpose, &tmp_sm);
            if (tmp_sm != NULL)
            {
                smap_destroy(tmp_sm);
                t2c_format_parse_error(block_parser, errUnexpectedAttribs, sec_purpose);
                return PR_SYNTAX_ERROR;
            }
            
            if (pr == PR_NOT_MATCHED)
            {
                break; // nothing more to do here
            }
        }
        
        if (pr != PR_MATCHED)
        {
            // some error occured
            return pr;
        }
        
        int nval = 0;        
        pr = t2c_parse_values_section(block_parser, 
            block_index, &test_index, abs_test_index, 
            &nval,
            tg, params, attr);
        
        if (pr != PR_PARSE_OK)
        {
            return pr;
        }
        
        if (counter == 1) // the first section
        {
            num_values = nval;
        }
        else
        {
            if (nval != num_values)
            {
                t2c_format_parse_error(block_parser, errDiffNumVals, NULL);
                return PR_SYNTAX_ERROR;
            }
        }
        
        t2c_parser_advance(block_parser);
        block_parser->beg = mist_find_in_range_first_not_of(block_parser->beg, block_parser->end, 
            wspace, sz_wspace);
    }
    while(block_parser->beg < block_parser->end);
    
    // if no sections specified, behave as if there was only one section 
    // and it was empty
    if (test_index == 0)
    {
        assert(num_values == 0);
               
        CGrowingArray records;
        if (grar_create(&records) == 0)
        {
            return PR_NO_MEMORY;
        }
        
        int bOK = t2c_parser_gen_test_sources(&records, 
            block_index, &test_index, abs_test_index,
            tg, params, attr);
        
        grar_destroy(&records);
        if (!bOK)
        {
            return PR_NO_MEMORY;
        }
    }

    return PR_PARSE_OK;
}

int
t2c_parser_gen_test_sources(CGrowingArray* records, 
    int block_index, int* test_index, int* abs_test_index,
    CMistTemplateGroup* tg,
    CStringMap* params, CStringMap* attr)
{
    assert(records != NULL);
    assert(block_index > 0);
    
    assert(test_index != NULL);
    assert(*test_index >= 0);
    
    assert(abs_test_index != NULL);
    assert(*abs_test_index >= 0);
    
    assert(tg != NULL);
    assert(params != NULL);
    
    size_t nrec = grar_get_size(records);
    if (nrec == 0)   // no values specified
    {
        CStringMap* local_params = smap_create();
        if (local_params == NULL)
        {
            return 0; // out of memory
        }
    
        ++(*test_index);
        ++(*abs_test_index);
        
        int bOK = t2c_generate_test_source(block_index, *test_index, *abs_test_index,
            tg, local_params, params, attr);
        
        smap_destroy(local_params);
        return bOK;
    }
    
    // nrec > 0 here
    CValueRecord** vr = grar_get_c_array(records, CValueRecord*);
    assert(vr != NULL);
    
    // array of current values
    const char** pvals = (const char**)malloc(sizeof(const char*) * nrec);
    if (pvals == NULL)
    {
        return 0;
    }
    memset(pvals, 0, sizeof(const char*) * nrec);
    
    int bOK = t2c_traverse_value_list(vr, pvals, 0, nrec, 
        block_index, test_index, abs_test_index,
        tg, params, attr);
    
    free(pvals);
    return bOK;
}

CValueElement*
t2c_vr_create_string_element(const char* beg, const char* end, long ntimes)
{
    assert(beg != NULL);
    assert(end != NULL);
    assert(beg <= end);
    assert(ntimes > 0);
    
    CValueElement* elem = (CValueElement*)malloc(sizeof(CValueElement));
    if (elem == NULL)
    {
        return NULL;
    }
    
    elem->str = mist_get_substring(beg, end);
    if (elem->str == NULL)
    {
        free(elem);
        return NULL;
    }
    
    elem->type = VRE_STRING;
    elem->ntimes = ntimes;
    
    elem->vmin = 0;
    elem->vmax = 0;
    
    return elem;
}

CValueElement*
t2c_vr_create_range_element(long vmin, long vmax)
{
    CValueElement* elem = (CValueElement*)malloc(sizeof(CValueElement));
    if (elem == NULL)
    {
        return NULL;
    }
    
    elem->str = NULL;
    elem->ntimes = 0;
    
    elem->type = VRE_RANGE;
    if (vmin <= vmax)
    {
        elem->vmin = vmin;
        elem->vmax = vmax;
    }
    else
    {
        elem->vmin = vmax;
        elem->vmax = vmin;
    }
    
    return elem;
}

void
t2c_vr_element_dtor(void* doomed, void* user_data)
{
    assert(doomed != NULL);
    
    CValueElement* ve = (CValueElement*)doomed;
    free(ve->str); // can be NULL but it's OK
    free(ve);
    
    return;
}

CValueRecord*
t2c_vr_create_value_record(enum EVRecordType type)
{
    CValueRecord* vr = (CValueRecord*)malloc(sizeof(CValueRecord));
    if (vr == NULL)
    {
        return NULL;
    }
    
    vr->type = type;
    if (grar_create(&(vr->elem)) == 0)
    {
        free(vr);
        return NULL;
    }
    
    return vr;
}

void 
t2c_vr_value_record_dtor(void* doomed, void* user_data)
{
    assert(doomed != NULL);
    
    CValueRecord* vr = (CValueRecord*)doomed;
    grar_destroy_with_elements(&(vr->elem), t2c_vr_element_dtor, NULL);
    free(vr);
    
    return;
}

EParserResult
t2c_vr_parse_value_record(CMistString* rstr, CValueRecord** rec, char** error_descr)
{
    assert(rstr != NULL);
    assert(rec != NULL);
    assert(error_descr != NULL);
    
    *error_descr = NULL;
    
    *rec = t2c_vr_create_value_record(VR_PLAIN);
    if (*rec == NULL)
    {
        return PR_NO_MEMORY;
    }
    
    CT2CParser vr_parser;
    t2c_parser_init(&vr_parser, rstr->str, rstr->str, rstr->str + strlen(rstr->str));
    
    // try to match SET(...) first
    EParserResult pr = t2c_vr_match_construct(&vr_parser, "SET");
    if (pr == PR_MATCHED)
    {
        (*rec)->type = VR_SET;
        pr = t2c_vr_parse_construct(&vr_parser, *rec);
        if (pr != PR_PARSE_OK)
        {
            t2c_vr_value_record_dtor(*rec, NULL);
            *rec = NULL;
            t2c_set_error(error_descr, &vr_parser);
            return pr;
        }
        return PR_PARSE_OK;
    }
    
    if (pr != PR_NOT_MATCHED)
    {
        t2c_vr_value_record_dtor(*rec, NULL);
        *rec = NULL;
        t2c_set_error(error_descr, &vr_parser);
        return pr;
    }
    
    // try to match RES(...) 
    pr = t2c_vr_match_construct(&vr_parser, "RES");
    if (pr == PR_MATCHED)
    {
        (*rec)->type = VR_RES;
        pr = t2c_vr_parse_construct(&vr_parser, *rec);
        if (pr != PR_PARSE_OK)
        {
            t2c_vr_value_record_dtor(*rec, NULL);
            *rec = NULL;
            t2c_set_error(error_descr, &vr_parser);
            return pr;
        }
        return PR_PARSE_OK;
    }
    
    if (pr != PR_NOT_MATCHED)
    {
        t2c_vr_value_record_dtor(*rec, NULL);
        *rec = NULL;
        t2c_set_error(error_descr, &vr_parser);
        return pr;
    }
    
    // neither SET(...) nor RES(...) - load the record as a plain one
    pr = t2c_vr_process_plain_record(&vr_parser, *rec);
    if (pr != PR_PARSE_OK)
    {
        t2c_vr_value_record_dtor(*rec, NULL);
        *rec = NULL;
        t2c_set_error(error_descr, &vr_parser);
        return pr;
    }

    return PR_PARSE_OK;
}

EParserResult
t2c_vr_match_construct(CT2CParser* vr_parser, const char* name)
{
    assert(t2c_is_parser_valid(vr_parser));
    assert(vr_parser->beg[0] != ' ' && vr_parser->beg[0] != '\t');
    assert(name != NULL);
    
    size_t clen = strlen(name);
    assert(clen > 0);
        
    static const char trail[] = "; \t"; // acceptable trailing characters
    static size_t sz_trail = sizeof(trail) / sizeof (trail[0]) - 1;
    
    const char* beg = vr_parser->beg;
    if ((size_t)(vr_parser->end - vr_parser->beg) < clen + 2 || // '+2' is for parentheses
        strncmp(beg, name, clen) != 0)
    {
        return PR_NOT_MATCHED;
    }
    
    beg = mist_find_in_range_first_not_of(beg + clen, vr_parser->end, ws, sz_ws);
    if (beg == vr_parser->end || *beg != '(')
    {
        t2c_format_parse_error_no_line(vr_parser, errMissingLeftParen, name);
        return PR_SYNTAX_ERROR;
    }
    
    const char* end = mist_find_in_range_last_not_of(beg, vr_parser->end, 
        trail, sz_trail);
    assert(end != vr_parser->end);
    assert(end >= beg);
    if (*end != ')')
    {
        t2c_format_parse_error_no_line(vr_parser, errMissingRightParen, name);
        return PR_SYNTAX_ERROR;
    }
    
    ++beg;
    // [beg; end) now denote what is within the parentheses
    
    vr_parser->beg = beg;
    vr_parser->end = end;
    
    return PR_MATCHED;
}

EParserResult
t2c_vr_parse_set_element(CT2CParser* vr_parser, CValueRecord* rec)
{
    assert(t2c_is_parser_valid(vr_parser));
    assert(rec != NULL);
    assert(rec->type == VR_SET);
    
    // we should have checked for empty elements earlier
    assert(vr_parser->beg < vr_parser->end);
    
    EParserResult pr = t2c_vr_match_range_element(vr_parser, rec);
    if (pr == PR_MATCHED)
    {
        return PR_PARSE_OK;
    }
    
    if (pr != PR_NOT_MATCHED)
    {
        // seems that some error has occured
        return pr;
    }
    
    CValueElement* ve = t2c_vr_create_string_element(vr_parser->beg, vr_parser->end, 1);
    if (ve == NULL)
    {
        return PR_NO_MEMORY;
    }
    
    if (grar_add_element(&(rec->elem), ve) == 0)
    {
        t2c_vr_element_dtor(ve, NULL);
        return PR_NO_MEMORY;
    }
    
    return PR_PARSE_OK;
}

EParserResult
t2c_vr_parse_res_element(CT2CParser* vr_parser, CValueRecord* rec)
{
    assert(t2c_is_parser_valid(vr_parser));
    assert(rec != NULL);
    assert(rec->type == VR_RES);
    
    // we should have checked for empty elements earlier
    assert(vr_parser->beg < vr_parser->end);
    
    static const char COLON = ':';
    CValueElement* ve = NULL;
    
    char* elem = mist_get_substring(vr_parser->beg, vr_parser->end);
    if (elem == NULL)
    {
        return PR_NO_MEMORY;
    }
    
    const char* pos = mist_find_in_range_last_not_of(vr_parser->beg, vr_parser->end, 
        ws_and_digits, sz_ws_and_digits);
        
    if (pos == vr_parser->end || *pos != COLON)
    {
        ve = t2c_vr_create_string_element(vr_parser->beg, vr_parser->end, 1);
    }
    else // a colon has been found, 'pos' points to it
    {
        const char* beg = mist_find_in_range_first_not_of(vr_parser->beg, pos, ws, sz_ws);
        if (beg == pos)
        {
            // only whitespace characters found before the colon
            t2c_format_parse_error_no_line(vr_parser, errInvalidRESElement, elem);
            free(elem);
            return PR_SYNTAX_ERROR;
        }

        const char* end = mist_find_in_range_last_not_of(beg, pos, ws, sz_ws);
        assert(end < pos);
        ++end;
        
        // [beg; end) now denotes the contents of the element;
        // extract the number of times to repeat the element
        pos = mist_find_in_range_first_not_of(pos + 1, vr_parser->end, ws, sz_ws);
        if (pos == vr_parser->end) 
        {
            // found only whitespace chars after the colon
            t2c_format_parse_error_no_line(vr_parser, errInvalidRESElement, elem);
            free(elem);
            return PR_SYNTAX_ERROR;
        }

        // try to convert the string to number
        char* res = NULL;
        
        errno = 0;
        long ntimes = strtol(&elem[pos - vr_parser->beg], &res, 10);
        assert(res != NULL);
        
        if (errno != 0 || *res != '\0')
        {
            t2c_format_parse_error_no_line(vr_parser, errConvertToNumFailed, &elem[pos - vr_parser->beg]);
            free(elem);
            return PR_SYNTAX_ERROR;
        }
        
        if (ntimes <= 0)
        {
            t2c_format_parse_error_no_line(vr_parser, errOnlyPosAllowed, NULL);
            free(elem);
            return PR_SYNTAX_ERROR;
        }
            
        ve = t2c_vr_create_string_element(beg, end, ntimes);
    }
    
    if (ve == NULL)
    {
        free(elem);
        return PR_NO_MEMORY;
    }
    
    if (grar_add_element(&(rec->elem), ve) == 0)
    {
        free(elem);
        t2c_vr_element_dtor(ve, NULL);
        return PR_NO_MEMORY;
    }
    
    free(elem);
    return PR_PARSE_OK;
}

EParserResult
t2c_vr_process_plain_record(CT2CParser* vr_parser, CValueRecord* rec)
{
    assert(t2c_is_parser_valid(vr_parser));
    assert(rec != NULL);
    assert(rec->type == VR_PLAIN);
    
    CValueElement* ve = t2c_vr_create_string_element(vr_parser->beg, vr_parser->end, 1);
    if (ve == NULL)
    {
        return PR_NO_MEMORY;
    }
    
    if (grar_add_element(&(rec->elem), ve) == 0)
    {
        t2c_vr_element_dtor(ve, NULL);
        return PR_NO_MEMORY;
    }
    
    return PR_PARSE_OK;
}

EParserResult
t2c_vr_parse_construct(CT2CParser* vr_parser, CValueRecord* rec)
{
    assert(t2c_is_parser_valid(vr_parser));
    assert(rec != NULL);
    assert(rec->type == VR_SET || rec->type == VR_RES);
    
    static const char QUOT  = '\"';
    static const char APOS  = '\'';
    static const char SEMI  =  ';';
    static const char SLASH = '\\';
    
    size_t len = vr_parser->end - vr_parser->beg;
    if (len == 0)
    {
        t2c_format_parse_error_no_line(vr_parser, errEmptyElement, 
            (rec->type == VR_SET ? "SET" : "RES"));
        return PR_SYNTAX_ERROR;
    }
    
    const char* pos = mist_find_in_range_first_not_of(vr_parser->beg, vr_parser->end, ws, sz_ws);
    if (pos == vr_parser->end)
    {
        t2c_format_parse_error_no_line(vr_parser, errEmptyElement, 
            (rec->type == VR_SET ? "SET" : "RES"));
        return PR_SYNTAX_ERROR;
    }
        
    CMistString* ms = mist_string_create("");
    if (ms == NULL)
    {
        return PR_NO_MEMORY;
    }
    
    if (mist_string_reserve(ms, len) != MIST_OK)
    {
        mist_string_destroy(ms);
        return PR_NO_MEMORY;
    }
    
    int in_quotes = 0;  // == 1 within double quotes ("...")
    int in_apos = 0;    // == 1 within apostrophed string ('...')
    
    while (1)
    {
        assert(!in_quotes);
        assert(!in_apos);
        
        if (pos == vr_parser->end || *pos == SEMI)
        {
            // trim the string, process the element, reset the string
            *(ms->str) = '\0';
            ms->str = ms->buf;
            mist_string_trim(ms);
            
            size_t tlen = strlen(ms->str);
            if (tlen == 0)
            {
                t2c_format_parse_error_no_line(vr_parser, errEmptyElement, 
                    (rec->type == VR_SET ? "SET" : "RES"));
                mist_string_destroy(ms);
                return PR_SYNTAX_ERROR;
            }
            
            CT2CParser el_parser;
            t2c_parser_init(&el_parser, ms->str, ms->str, ms->str + tlen);
            
            EParserResult pr;
            if (rec->type == VR_SET)
            {
                pr = t2c_vr_parse_set_element(&el_parser, rec);
            }
            else // RES-construct
            {
                pr = t2c_vr_parse_res_element(&el_parser, rec);
            }
            
            if (pr != PR_PARSE_OK)
            {
                t2c_set_error(&(vr_parser->error), &el_parser);
                mist_string_destroy(ms);
                return pr;
            }
            
            if (pos == vr_parser->end)
            {
                break;
            }
            else
            {
                ++pos;
                mist_string_reset(ms);
            }
        }
        else if (*pos == QUOT || *pos == APOS)
        {
            char ch = *pos;
            *(ms->str) = ch;
            ++(ms->str);
            ++pos;
            
            const char* beg = pos;
            
            if (ch == QUOT)
            {
                in_quotes = 1;
            }
            else
            {
                in_apos = 1;
            }
            
            while (pos < vr_parser->end)
            {
                if (*pos == SLASH)
                {
                    if (mist_is_char_escaped(pos, beg))
                    {
                        // we have already copied the slash to 'ms'
                        beg = pos + 1;
                    }
                    else
                    {
                        *(ms->str) = *pos;
                        ++(ms->str);
                        beg = pos;
                    }
                }
                else if (*pos == ch)
                {
                    if (mist_is_char_escaped(pos, beg))
                    {
                        assert(ms->str >= ms->buf + 1);
                        *(ms->str - 1) = *pos;
                    }
                    else
                    {
                        *(ms->str) = *pos;
                        ++(ms->str);
                        
                        if (ch == QUOT)
                        {
                            in_quotes = 0;
                            ++pos;
                            break;
                        }
                        else
                        {
                            in_apos = 0;
                            ++pos;
                            break;
                        }
                    }
                }
                else
                {
                    *(ms->str) = *pos;
                    ++(ms->str);
                }
                ++pos;
            }
            
            if (in_quotes)
            {
                t2c_format_parse_error_no_line(vr_parser, errLoneQuot, 
                    (rec->type == VR_SET ? "SET" : "RES"));
                mist_string_destroy(ms);
                return PR_SYNTAX_ERROR;
            }
            
            if (in_apos)
            {
                t2c_format_parse_error_no_line(vr_parser, errLoneApos, 
                    (rec->type == VR_SET ? "SET" : "RES"));
                mist_string_destroy(ms);
                return PR_SYNTAX_ERROR;
            }
        }
        else
        {
            *(ms->str) = *pos;
            ++(ms->str);
            ++pos;
        }
    }
    
    mist_string_destroy(ms);
    return PR_PARSE_OK;
}

EParserResult
t2c_vr_match_range_element(CT2CParser* vr_parser, CValueRecord* rec)
{
    assert(t2c_is_parser_valid(vr_parser));
    assert(rec != NULL);
    assert(rec->type == VR_SET);
    
    // we should have checked for empty elements earlier
    assert(vr_parser->beg < vr_parser->end);
    
    static const char DOT = '.';
    
    const char* first_end = mist_find_in_range_first_not_of(vr_parser->beg, vr_parser->end, 
        ws_and_digits, sz_ws_and_digits);
    if (first_end == vr_parser->end || *first_end != DOT)
    {
        return PR_NOT_MATCHED;
    }
    
    const char* last_beg = mist_find_in_range_last_not_of(vr_parser->beg, vr_parser->end, 
        ws_and_digits, sz_ws_and_digits);
    assert(last_beg != vr_parser->end);
    if (*last_beg != DOT || last_beg != first_end + 1)
    {
        return PR_NOT_MATCHED;
    }
    ++last_beg;
    
    CMistString* sfirst = mist_string_create_from_range(vr_parser->beg, first_end);
    if (sfirst == NULL)
    {
        return PR_NO_MEMORY;
    }
    
    CMistString* slast = mist_string_create_from_range(last_beg, vr_parser->end);
    if (slast == NULL)
    {
        mist_string_destroy(sfirst);
        return PR_NO_MEMORY;
    }
    
    mist_string_trim(sfirst);
    mist_string_trim(slast);
    
    if (sfirst->str[0] == '\0' || slast->str[0] == '\0')
    {
        mist_string_destroy(sfirst);
        mist_string_destroy(slast);
        
        char* elem = mist_get_substring(vr_parser->beg, vr_parser->end);
        if (elem == NULL)
        {
            return PR_NO_MEMORY;
        }
        
        t2c_format_parse_error_no_line(vr_parser, errInvalidSETElement, elem);
        free(elem);
        return PR_SYNTAX_ERROR;
    }
    
    char* res = NULL;
    errno = 0;
    long vmin = strtol(sfirst->str, &res, 10);
    assert(res != NULL);
        
    if (errno != 0 || *res != '\0')
    {
        t2c_format_parse_error_no_line(vr_parser, errConvertToNumFailed, sfirst->str);
        mist_string_destroy(sfirst);
        mist_string_destroy(slast);
        return PR_SYNTAX_ERROR;
    }
    
    res = NULL;
    errno = 0;
    long vmax = strtol(slast->str, &res, 10);
    assert(res != NULL);
        
    if (errno != 0 || *res != '\0')
    {
        t2c_format_parse_error_no_line(vr_parser, errConvertToNumFailed, slast->str);
        mist_string_destroy(sfirst);
        mist_string_destroy(slast);
        return PR_SYNTAX_ERROR;
    }
    
    mist_string_destroy(sfirst);
    mist_string_destroy(slast);
    
    CValueElement* ve = t2c_vr_create_range_element(vmin, vmax);
    if (ve == NULL)
    {
        return PR_NO_MEMORY;
    }
    
    if (grar_add_element(&(rec->elem), ve) == 0)
    {
        t2c_vr_element_dtor(ve, NULL);
        return PR_NO_MEMORY;
    }
    
    return PR_MATCHED;
}

int
t2c_generate_test_source(int block_index, int test_index, int abs_test_index,
    CMistTemplateGroup* tg, 
    CStringMap* local_params, CStringMap* global_params, CStringMap* attr)
{
    assert(tg != NULL);
    assert(local_params != NULL);
    assert(global_params != NULL);
    
    assert(block_index >= 1);
    assert(test_index >= 1);
    assert(abs_test_index >= 1);
    
    // set values of global parameters: "block_index", "test_index", "abs_test_index" 
    if (t2c_set_index_values(global_params, block_index, test_index, abs_test_index) == 0)
    {
        return 0;
    }
    
    // set values of local parameters of the test case: 
    // "block_index", "test_index", "abs_test_index" 
    if (t2c_set_index_values(local_params, block_index, test_index, abs_test_index) == 0)
    {
        return 0;
    }
    
    // generate the source of the test
    mist_tg_clear_values_impl(tg);
    if (mist_tg_set_values_impl(tg, local_params) != MIST_OK)
    {
        return 0;
    }
    
    if (attr != NULL && 
        mist_tg_set_values_impl(tg, attr) != MIST_OK)
    {
        return 0;
    }

    CGrowingArray* vals = mist_tg_evaluate_impl(tg);
    if (vals == NULL)
    {
        return 0;
    }
    assert(grar_get_size(vals) == 1);

    const char* source = grar_get_element(vals, const char*, 0);
    assert(source != NULL);
    
    if (smap_add_element(global_params, "test_case", source) == 0)
    {
        return 0;
    }
    
    return 1;
}

int
t2c_traverse_value_list(CValueRecord** vr, const char** pvals, size_t start, size_t nvals,
    int block_index, int* test_index, int* abs_test_index,
    CMistTemplateGroup* tg, CStringMap* global_params, CStringMap* attr)
{
    assert(tg != NULL);
    assert(global_params != NULL);
    
    assert(block_index > 0);
    assert(test_index != NULL);
    assert(*test_index >= 0);
    assert(abs_test_index != NULL);
    assert(*abs_test_index >= 0);
    
    assert(vr != NULL);
    assert(pvals != NULL);
    assert(start <= nvals);
    
    // Current number of the test case defined by this <VALUES>/<PURPOSE> section.
    // It does NOT equal '*test_index'.
    static long tc_num = 0; 
    if (start == 0)
    {
        tc_num = 0; // reset the counter
    }
    
    if (start == nvals) // a special case
    {
        ++tc_num;
        
        // set the values of the parameters and generate the tests
        ++(*test_index);
        ++(*abs_test_index);
        
        CStringMap* local_params = smap_create();
        if (local_params == NULL)
        {
            return 0;
        }
        
        // Values of VR_PLAIN and VR_SET records are already set in 'pvals'
        // Now to process VR_RES ones:
        for (size_t i = 0; i < nvals; ++i)
        {
            if (vr[i]->type == VR_RES)
            {
                CValueElement* ve = t2c_find_res_element(vr[i], tc_num);
                assert(ve != NULL);
                assert(ve->str != NULL);
                
                pvals[i] = ve->str;
            }
        }
        
        if (t2c_set_numbered_params(pvals, nvals, local_params) == 0)
        {
            smap_destroy(local_params);
            return 0;
        }
        
        int bOK = t2c_generate_test_source(block_index, *test_index, *abs_test_index,
            tg, local_params, global_params, attr);
        
        smap_destroy(local_params);
        return bOK;
    }
    
    size_t nelem = grar_get_size(&(vr[start]->elem));
    CValueElement* el = NULL;
    
    switch (vr[start]->type)
    {
    case VR_PLAIN:
        assert(nelem == 1);
        
        el = grar_get_element(&(vr[start]->elem), CValueElement*, 0);
        assert(el != NULL);
        assert(el->str != NULL);
        assert(el->type == VRE_STRING);
        
        pvals[start] = el->str;
        
        if (t2c_traverse_value_list(vr, pvals, start + 1, nvals,
                block_index, test_index, abs_test_index,
                tg, global_params, attr) == 0)
        {
            return 0;
        }
        break;
        
    case VR_SET:
        {
            for (size_t i = 0; i < nelem; ++i)  // for each element ...
            {
                el = grar_get_element(&(vr[start]->elem), CValueElement*, i);
                assert(el != NULL);
                
                if (el->type == VRE_STRING)
                {
                    assert(el->str != NULL);
                    pvals[start] = el->str;
                    
                    if (t2c_traverse_value_list(vr, pvals, start + 1, nvals,
                            block_index, test_index, abs_test_index,
                            tg, global_params, attr) == 0)
                    {
                        return 0;
                    }
                }
                else
                {
                    assert(el->type == VRE_RANGE);
                    assert(el->vmin <= el->vmax);
                    
                    char snum[MAX_NUM_DIGITS + 1];  // for numbers
                    for (long n = el->vmin; n <= el->vmax; ++n)
                    {
                        sprintf(&snum[0], "%ld", n);
                        pvals[start] = &snum[0];
                    
                        if (t2c_traverse_value_list(vr, pvals, start + 1, nvals,
                                block_index, test_index, abs_test_index,
                                tg, global_params, attr) == 0)
                        {
                            return 0;
                        }
                    } // for n
                }
            } // for i
        }
        
        break;
        
    case VR_RES:
        // We do not set the value here: test case number is yet unknown
        // This will be done when start == nvals.
        if (t2c_traverse_value_list(vr, pvals, start + 1, nvals,
                block_index, test_index, abs_test_index,
                tg, global_params, attr) == 0)
        {
            return 0;
        }
        break;
        
    default:
        assert(0 && "Unknown type of a value record");
    }

    return 1;
}

int 
t2c_set_numbered_params(const char** pvals, size_t nvals, CStringMap* local_params)
{
    assert(pvals != NULL);
    assert(nvals > 0);
    assert(local_params != NULL);
  
    char snum[MAX_NUM_DIGITS + 1];  // for numbers
    
    for (int i = 0; i < (int)(nvals); ++i)
    {
        assert(pvals[i] != NULL);
        sprintf(&snum[0], "%d", i);
        
        if (smap_add_element(local_params, &snum[0], pvals[i]) == 0 ||
            smap_add_element(local_params, "param", pvals[i]) == 0)
        {
            return 0;
        }
    }

    return 1;    
}

CValueElement*
t2c_find_res_element(CValueRecord* vr, long tc_num)
{
    assert(vr != NULL);
    assert(vr->type == VR_RES);
    
    size_t nelems = grar_get_size(&(vr->elem));
    assert(nelems > 0); // empty RES-constructs are not allowed
    
    CValueElement** el = grar_get_c_array(&(vr->elem), CValueElement*);
    assert(el != NULL);
    
    long accum = 0;
    for (long i = 0; i < (long)nelems; ++i)
    {
        accum += el[i]->ntimes;
        if (accum >= tc_num)
        {
            return el[i];
        }
    }
    
    return el[nelems - 1];
}

int
print_data_dir()
{
    char* datadir = get_data_dir();
    if (datadir == NULL)
    {
        return 0;
    }
    
    printf("%s\n", datadir);
    free(datadir);
    return 1;
}

int 
populate_test_module_list(struct SSettings* psettings)
{
    assert(psettings != NULL);
    assert(psettings->test_group_params != NULL);
    
    static char test_module_key[] = "TEST_MODULE";
    
    if (psettings->test_name_file_path != NULL &&
        mist_file_exists(psettings->test_name_file_path))
    {
        char* data = NULL;
        if (mist_file_read_all(psettings->test_name_file_path, &data) != MIST_OK)
        {
            // if loading of the file fails, it is acceptable,
            // just do nothing.
            return 1;
        }
        
        int bOK = 1;
        
        const char* beg = data;
        const char* end = data + strlen(data);
        
        beg = mist_find_in_range_first_not_of(beg, end, wspace, sz_wspace);
        while (beg < end)
        {
            const char* record_end = mist_find_in_range_first_of(beg, end, wspace, sz_wspace);
            char c = *record_end;
            
            data[record_end - (const char*)data] = '\0';
            bOK = smap_add_element(psettings->test_group_params, &test_module_key[0], beg);
            data[record_end - (const char*)data] = c;
            
            if (!bOK)
            {
                break;
            }
            
            beg = mist_find_in_range_first_not_of(record_end, end, wspace, sz_wspace);
        }
        
        free(data);
        
        if (!bOK)
        {
            return 0;
        }
    }
    return 1;
}

int 
print_error(const char* format, ...)
{
    assert(format != NULL);
    
    int res = -1;
    va_list ap;
    
    // output the prefix
    res = fprintf(stderr, "%s: ", T2C_SHORT_NAME);
    if (res < 0)
    {
        return res;
    }

    va_start(ap, format);
    res = vfprintf(stderr, format, ap);
    va_end(ap);
    
    if (res < 0)
    {
        return res;
    }
    
    res = fprintf(stderr, "\n");
    return res;
}
///////////////////////////////////////////////////////////////////////////
